# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list


def test_ByteMaskedArray():
    content = ak.operations.from_iter(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    mask = ak.index.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    array = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    assert to_list(array) == [[0.0, 1.1, 2.2], [], None, None, [6.6, 7.7, 8.8, 9.9]]
    assert to_list(array[-1]) == [6.6, 7.7, 8.8, 9.9]
    assert to_list(array[-2]) is None
    assert to_list(array[1:]) == [[], None, None, [6.6, 7.7, 8.8, 9.9]]
    assert array[4, 1] == 7.7
    assert to_list(array[2:, 1]) == [None, None, 7.7]
    assert to_list(array[2:, [2, 1, 1, 0]]) == [None, None, [8.8, 7.7, 7.7, 6.6]]

    content = ak.operations.from_iter(
        [
            {"x": 0.0, "y": []},
            {"x": 1.1, "y": [1]},
            {"x": 2.2, "y": [2, 2]},
            {"x": 3.3, "y": [3, 3, 3]},
            {"x": 4.4, "y": [4, 4, 4, 4]},
        ],
        highlevel=False,
    )
    mask = ak.index.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    array = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    assert to_list(array) == [
        {"x": 0.0, "y": []},
        {"x": 1.1, "y": [1]},
        None,
        None,
        {"x": 4.4, "y": [4, 4, 4, 4]},
    ]
    assert to_list(array["x"]) == [0.0, 1.1, None, None, 4.4]
    assert to_list(array[["x", "y"]]) == [
        {"x": 0.0, "y": []},
        {"x": 1.1, "y": [1]},
        None,
        None,
        {"x": 4.4, "y": [4, 4, 4, 4]},
    ]


def test_ByteMaskedArray_jaggedslice0():
    array = ak.operations.from_iter(
        [[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    index = ak.index.Index64(np.array([0, 1, 2, 3], dtype=np.int64))
    indexedarray = ak.contents.IndexedOptionArray(index, array)
    assert to_list(indexedarray) == [
        [0.0, 1.1, 2.2],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert to_list(indexedarray[ak.highlevel.Array([[0, -1], [0], [], [1, 1]])]) == [
        [0.0, 2.2],
        [3.3],
        [],
        [7.7, 7.7],
    ]

    mask = ak.index.Index8(np.array([0, 0, 0, 0], dtype=np.int8))
    maskedarray = ak.contents.ByteMaskedArray(mask, array, valid_when=False)
    assert to_list(maskedarray) == [
        [0.0, 1.1, 2.2],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert to_list(maskedarray[ak.highlevel.Array([[0, -1], [0], [], [1, 1]])]) == [
        [0.0, 2.2],
        [3.3],
        [],
        [7.7, 7.7],
    ]


def test_ByteMaskedArray_jaggedslice1():
    model = ak.highlevel.Array(
        [[0.0, 1.1, None, 2.2], [], [3.3, None, 4.4], [5.5], [6.6, 7.7, None, 8.8, 9.9]]
    )
    assert to_list(
        model[ak.highlevel.Array([[3, 2, 1, 1, 0], [], [1], [0, 0], [1, 2]])]
    ) == [
        [2.2, None, 1.1, 1.1, 0.0],
        [],
        [None],
        [5.5, 5.5],
        [7.7, None],
    ]

    content = ak.contents.NumpyArray(
        np.array([0.0, 1.1, 999, 2.2, 3.3, 123, 4.4, 5.5, 6.6, 7.7, 321, 8.8, 9.9])
    )
    mask = ak.index.Index8(
        np.array([0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0], dtype=np.int8)
    )
    maskedarray = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    offsets = ak.index.Index64(np.array([0, 4, 4, 7, 8, 13], dtype=np.int64))
    listarray = ak.highlevel.Array(ak.contents.ListOffsetArray(offsets, maskedarray))
    assert to_list(listarray) == to_list(model)
    assert to_list(
        listarray[ak.highlevel.Array([[3, 2, 1, 1, 0], [], [1], [0, 0], [1, 2]])]
    ) == [[2.2, None, 1.1, 1.1, 0.0], [], [None], [5.5, 5.5], [7.7, None]]


def test_ByteMaskedArray_jaggedslice2():
    model = ak.highlevel.Array(
        [
            [[0.0, 1.1, None, 2.2], [], [3.3, None, 4.4]],
            [],
            [[5.5]],
            [[6.6, 7.7, None, 8.8, 9.9]],
        ]
    )
    assert to_list(
        model[ak.highlevel.Array([[[3, 2, 1, 1, 0], [], [1]], [], [[0, 0]], [[1, 2]]])]
    ) == [[[2.2, None, 1.1, 1.1, 0.0], [], [None]], [], [[5.5, 5.5]], [[7.7, None]]]

    content = ak.contents.NumpyArray(
        np.array([0.0, 1.1, 999, 2.2, 3.3, 123, 4.4, 5.5, 6.6, 7.7, 321, 8.8, 9.9])
    )
    mask = ak.index.Index8(
        np.array([0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0], dtype=np.int8)
    )
    maskedarray = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    offsets = ak.index.Index64(np.array([0, 4, 4, 7, 8, 13], dtype=np.int64))
    sublistarray = ak.contents.ListOffsetArray(offsets, maskedarray)
    offsets2 = ak.index.Index64(np.array([0, 3, 3, 4, 5], dtype=np.int64))
    listarray = ak.highlevel.Array(ak.contents.ListOffsetArray(offsets2, sublistarray))
    assert to_list(listarray) == to_list(model)
    assert to_list(
        listarray[
            ak.highlevel.Array([[[3, 2, 1, 1, 0], [], [1]], [], [[0, 0]], [[1, 2]]])
        ]
    ) == [[[2.2, None, 1.1, 1.1, 0.0], [], [None]], [], [[5.5, 5.5]], [[7.7, None]]]


def test_ByteMaskedArray_jaggedslice3():
    model = ak.highlevel.Array(
        [
            [[[0.0, 1.1, None, 2.2], [], [3.3, None, 4.4]], []],
            [[[5.5]], [[6.6, 7.7, None, 8.8, 9.9]]],
        ]
    )
    assert to_list(
        model[
            ak.highlevel.Array([[[[3, 2, 1, 1, 0], [], [1]], []], [[[0, 0]], [[1, 2]]]])
        ]
    ) == [[[[2.2, None, 1.1, 1.1, 0.0], [], [None]], []], [[[5.5, 5.5]], [[7.7, None]]]]

    content = ak.contents.NumpyArray(
        np.array([0.0, 1.1, 999, 2.2, 3.3, 123, 4.4, 5.5, 6.6, 7.7, 321, 8.8, 9.9])
    )
    mask = ak.index.Index8(
        np.array([0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0], dtype=np.int8)
    )
    maskedarray = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    offsets = ak.index.Index64(np.array([0, 4, 4, 7, 8, 13], dtype=np.int64))
    subsublistarray = ak.contents.ListOffsetArray(offsets, maskedarray)
    offsets2 = ak.index.Index64(np.array([0, 3, 3, 4, 5], dtype=np.int64))
    sublistarray = ak.contents.ListOffsetArray(offsets2, subsublistarray)
    offsets3 = ak.index.Index64(np.array([0, 2, 4], dtype=np.int64))
    listarray = ak.highlevel.Array(ak.contents.ListOffsetArray(offsets3, sublistarray))
    assert to_list(listarray) == to_list(model)
    assert to_list(
        listarray[
            ak.highlevel.Array([[[[3, 2, 1, 1, 0], [], [1]], []], [[[0, 0]], [[1, 2]]]])
        ]
    ) == [[[[2.2, None, 1.1, 1.1, 0.0], [], [None]], []], [[[5.5, 5.5]], [[7.7, None]]]]


def test_ByteMaskedArray_to_slice():
    content = ak.contents.NumpyArray(
        np.array([5, 2, 999, 3, 9, 123, 1], dtype=np.int64)
    )
    mask = ak.index.Index8(np.array([0, 0, 1, 0, 0, 1, 0], dtype=np.int8))
    maskedarray = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    assert to_list(maskedarray) == [5, 2, None, 3, 9, None, 1]


def test_ByteMaskedArray_as_slice():
    array = ak.highlevel.Array(
        [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], check_valid=True
    )
    slicecontent = ak.operations.from_iter([3, 6, 999, 123, -2, 6], highlevel=False)
    slicemask = ak.index.Index8(np.array([0, 0, 1, 1, 0, 0], dtype=np.int8))
    slicearray = ak.contents.ByteMaskedArray(slicemask, slicecontent, valid_when=False)

    assert to_list(array[slicearray]) == [3.3, 6.6, None, None, 8.8, 6.6]


def test_ByteMaskedArray_whatever():
    content = ak.operations.from_iter(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9], [321]],
        highlevel=False,
    )
    mask = ak.index.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    array = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    assert to_list(array) == [[0.0, 1.1, 2.2], [], None, None, [6.6, 7.7, 8.8, 9.9]]

    assert ak.operations.is_valid(ak.highlevel.Array(array))


def test_ByteMaskedArray_num():
    content = ak.operations.from_iter(
        [
            [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
            [],
            [[5.5]],
            [[6.6, 7.7, 8.8, 9.9]],
            [[], [10.0, 11.1, 12.2]],
        ],
        highlevel=False,
    )
    mask = ak.index.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    array = ak.highlevel.Array(
        ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    )
    assert to_list(array) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
    ]
    assert ak.operations.num(array, axis=0) == 5
    assert ak.operations.num(array, axis=-3) == 5
    assert to_list(ak.operations.num(array, axis=1)) == [
        3,
        0,
        None,
        None,
        2,
    ]
    assert to_list(ak.operations.num(array, axis=-2)) == [
        3,
        0,
        None,
        None,
        2,
    ]
    assert to_list(ak.operations.num(array, axis=2)) == [
        [3, 0, 2],
        [],
        None,
        None,
        [0, 3],
    ]
    assert to_list(ak.operations.num(array, axis=-1)) == [
        [3, 0, 2],
        [],
        None,
        None,
        [0, 3],
    ]


def test_ByteMaskedArray_flatten():
    content = ak.operations.from_iter(
        [
            [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
            [],
            [[5.5]],
            [[6.6, 7.7, 8.8, 9.9]],
            [[], [10.0, 11.1, 12.2]],
        ],
        highlevel=False,
    )
    mask = ak.index.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    array = ak.highlevel.Array(
        ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    )
    assert to_list(array) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
    ]
    assert to_list(ak.operations.flatten(array, axis=1)) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [],
        [10.0, 11.1, 12.2],
    ]
    assert to_list(ak.operations.flatten(array, axis=-2)) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [],
        [10.0, 11.1, 12.2],
    ]
    assert to_list(ak.operations.flatten(array, axis=2)) == [
        [0.0, 1.1, 2.2, 3.3, 4.4],
        [],
        None,
        None,
        [10.0, 11.1, 12.2],
    ]
    assert to_list(ak.operations.flatten(array, axis=-1)) == [
        [0.0, 1.1, 2.2, 3.3, 4.4],
        [],
        None,
        None,
        [10.0, 11.1, 12.2],
    ]


def test_IndexedOptionArray_pad_none():
    array = ak.highlevel.Array(
        [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None, None, [[], [10.0, 11.1, 12.2]]]
    )
    assert to_list(array) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
    ]
    assert to_list(ak.operations.pad_none(array, 7, axis=0)) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
        None,
        None,
    ]
    assert to_list(ak.operations.pad_none(array, 7, axis=-3)) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
        None,
        None,
    ]
    assert to_list(ak.operations.pad_none(array, 3, axis=1)) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [None, None, None],
        None,
        None,
        [[], [10.0, 11.1, 12.2], None],
    ]
    assert to_list(ak.operations.pad_none(array, 3, axis=-2)) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [None, None, None],
        None,
        None,
        [[], [10.0, 11.1, 12.2], None],
    ]
    assert to_list(ak.operations.pad_none(array, 3, axis=2)) == [
        [[0.0, 1.1, 2.2], [None, None, None], [3.3, 4.4, None]],
        [],
        None,
        None,
        [[None, None, None], [10.0, 11.1, 12.2]],
    ]
    assert to_list(ak.operations.pad_none(array, 3, axis=-1)) == [
        [[0.0, 1.1, 2.2], [None, None, None], [3.3, 4.4, None]],
        [],
        None,
        None,
        [[None, None, None], [10.0, 11.1, 12.2]],
    ]
    assert to_list(ak.operations.pad_none(array, 3, axis=0, clip=True)) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
    ]
    assert to_list(ak.operations.pad_none(array, 3, axis=-3, clip=True)) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
    ]
    assert to_list(ak.operations.pad_none(array, 2, axis=1, clip=True)) == [
        [[0.0, 1.1, 2.2], []],
        [None, None],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
    ]
    assert to_list(ak.operations.pad_none(array, 2, axis=-2, clip=True)) == [
        [[0.0, 1.1, 2.2], []],
        [None, None],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
    ]
    assert to_list(ak.operations.pad_none(array, 2, axis=2, clip=True)) == [
        [[0.0, 1.1], [None, None], [3.3, 4.4]],
        [],
        None,
        None,
        [[None, None], [10.0, 11.1]],
    ]
    assert to_list(ak.operations.pad_none(array, 2, axis=-1, clip=True)) == [
        [[0.0, 1.1], [None, None], [3.3, 4.4]],
        [],
        None,
        None,
        [[None, None], [10.0, 11.1]],
    ]


def test_ByteMaskedArray_pad_none():
    content = ak.operations.from_iter(
        [
            [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
            [],
            [[5.5]],
            [[6.6, 7.7, 8.8, 9.9]],
            [[], [10.0, 11.1, 12.2]],
        ],
        highlevel=False,
    )
    mask = ak.index.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    array = ak.highlevel.Array(
        ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    )
    assert to_list(array) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
    ]
    assert to_list(ak.operations.pad_none(array, 7, axis=0)) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
        None,
        None,
    ]
    assert to_list(ak.operations.pad_none(array, 7, axis=-3)) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
        None,
        None,
    ]
    assert to_list(ak.operations.pad_none(array, 3, axis=1)) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [None, None, None],
        None,
        None,
        [[], [10.0, 11.1, 12.2], None],
    ]
    assert to_list(ak.operations.pad_none(array, 3, axis=-2)) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [None, None, None],
        None,
        None,
        [[], [10.0, 11.1, 12.2], None],
    ]
    assert to_list(ak.operations.pad_none(array, 3, axis=2)) == [
        [[0.0, 1.1, 2.2], [None, None, None], [3.3, 4.4, None]],
        [],
        None,
        None,
        [[None, None, None], [10.0, 11.1, 12.2]],
    ]
    assert to_list(ak.operations.pad_none(array, 3, axis=-1)) == [
        [[0.0, 1.1, 2.2], [None, None, None], [3.3, 4.4, None]],
        [],
        None,
        None,
        [[None, None, None], [10.0, 11.1, 12.2]],
    ]
    assert to_list(ak.operations.pad_none(array, 3, axis=0, clip=True)) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
    ]
    assert to_list(ak.operations.pad_none(array, 3, axis=-3, clip=True)) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
    ]
    assert to_list(ak.operations.pad_none(array, 2, axis=1, clip=True)) == [
        [[0.0, 1.1, 2.2], []],
        [None, None],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
    ]
    assert to_list(ak.operations.pad_none(array, 2, axis=-2, clip=True)) == [
        [[0.0, 1.1, 2.2], []],
        [None, None],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
    ]
    assert to_list(ak.operations.pad_none(array, 2, axis=2, clip=True)) == [
        [[0.0, 1.1], [None, None], [3.3, 4.4]],
        [],
        None,
        None,
        [[None, None], [10.0, 11.1]],
    ]
    assert to_list(ak.operations.pad_none(array, 2, axis=-1, clip=True)) == [
        [[0.0, 1.1], [None, None], [3.3, 4.4]],
        [],
        None,
        None,
        [[None, None], [10.0, 11.1]],
    ]


def test_ByteMaskedArray_reduce():
    content = ak.contents.NumpyArray(
        np.array(
            [
                2,
                3,
                5,
                7,
                11,
                0,
                0,
                0,
                0,
                0,
                31,
                37,
                41,
                43,
                47,
                53,
                59,
                61,
                67,
                71,
                0,
                0,
                0,
                0,
                0,
                101,
                103,
                107,
                109,
                113,
            ],
            dtype=np.int64,
        )
    )
    offsets1 = ak.index.Index64(np.array([0, 5, 10, 15, 20, 25, 30], dtype=np.int64))
    listoffsetarray = ak.contents.ListOffsetArray(offsets1, content)
    # index = ak.index.Index64(np.array([0, -1, 2, 3, -1, 5], dtype=np.int64))
    # optionarray = ak.contents.IndexedOptionArray(index, listoffsetarray)
    mask = ak.index.Index8(np.array([0, 1, 0, 0, 1, 0], dtype=np.int8))
    optionarray = ak.contents.ByteMaskedArray(mask, listoffsetarray, valid_when=False)
    offsets2 = ak.index.Index64(np.array([0, 3, 6], dtype=np.int64))
    depth2 = ak.contents.ListOffsetArray(offsets2, optionarray)
    assert to_list(depth2) == [
        [[2, 3, 5, 7, 11], None, [31, 37, 41, 43, 47]],
        [[53, 59, 61, 67, 71], None, [101, 103, 107, 109, 113]],
    ]

    assert to_list(ak.prod(depth2, -1, highlevel=False)) == [
        [2 * 3 * 5 * 7 * 11, None, 31 * 37 * 41 * 43 * 47],
        [53 * 59 * 61 * 67 * 71, None, 101 * 103 * 107 * 109 * 113],
    ]

    assert to_list(ak.prod(depth2, -2, highlevel=False)) == [
        [31 * 2, 37 * 3, 41 * 5, 43 * 7, 47 * 11],
        [101 * 53, 103 * 59, 107 * 61, 109 * 67, 113 * 71],
    ]

    assert to_list(ak.prod(depth2, -3, highlevel=False)) == [
        [53 * 2, 59 * 3, 61 * 5, 67 * 7, 71 * 11],
        [],
        [101 * 31, 103 * 37, 107 * 41, 109 * 43, 113 * 47],
    ]

    content = ak.contents.NumpyArray(
        np.array(
            [
                2,
                3,
                5,
                7,
                11,
                0,
                0,
                0,
                0,
                0,
                31,
                37,
                41,
                43,
                47,
                53,
                59,
                61,
                67,
                71,
                0,
                0,
                0,
                0,
                0,
                101,
                103,
                107,
                109,
                113,
            ],
            dtype=np.int64,
        )
    )
    # index = ak.index.Index64(np.array([0, 1, 2, 3, 4, -1, -1, -1, -1, -1, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, -1, -1, -1, -1, -1, 25, 26, 27, 28, 29], dtype=np.int64))
    # optionarray = ak.contents.IndexedOptionArray(index, content)
    mask = ak.index.Index8(
        np.array(
            [
                0,
                0,
                0,
                0,
                0,
                1,
                1,
                1,
                1,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                1,
                1,
                1,
                1,
                0,
                0,
                0,
                0,
                0,
            ],
            dtype=np.int8,
        )
    )
    optionarray = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    offsets1 = ak.index.Index64(np.array([0, 5, 10, 15, 20, 25, 30], dtype=np.int64))
    listoffsetarray = ak.contents.ListOffsetArray(offsets1, optionarray)
    offsets2 = ak.index.Index64(np.array([0, 3, 6], dtype=np.int64))
    depth2 = ak.contents.ListOffsetArray(offsets2, listoffsetarray)
    assert to_list(depth2) == [
        [[2, 3, 5, 7, 11], [None, None, None, None, None], [31, 37, 41, 43, 47]],
        [
            [53, 59, 61, 67, 71],
            [None, None, None, None, None],
            [101, 103, 107, 109, 113],
        ],
    ]

    assert to_list(ak.prod(depth2, -1, highlevel=False)) == [
        [2 * 3 * 5 * 7 * 11, 1 * 1 * 1 * 1 * 1, 31 * 37 * 41 * 43 * 47],
        [53 * 59 * 61 * 67 * 71, 1 * 1 * 1 * 1 * 1, 101 * 103 * 107 * 109 * 113],
    ]

    assert to_list(ak.prod(depth2, -2, highlevel=False)) == [
        [31 * 2, 37 * 3, 41 * 5, 43 * 7, 47 * 11],
        [101 * 53, 103 * 59, 107 * 61, 109 * 67, 113 * 71],
    ]

    assert to_list(ak.prod(depth2, -3, highlevel=False)) == [
        [53 * 2, 59 * 3, 61 * 5, 67 * 7, 71 * 11],
        [1, 1, 1, 1, 1],
        [101 * 31, 103 * 37, 107 * 41, 109 * 43, 113 * 47],
    ]


def test_ByteMaskedArray_localindex():
    content = ak.operations.from_iter(
        [
            [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
            [],
            [[5.5]],
            [[6.6, 7.7, 8.8, 9.9]],
            [[], [10.0, 11.1, 12.2]],
        ],
        highlevel=False,
    )
    mask = ak.index.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    array = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    assert to_list(array) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
    ]
    assert to_list(ak._do.local_index(array, axis=0)) == [0, 1, 2, 3, 4]
    assert to_list(ak._do.local_index(array, axis=-3)) == [0, 1, 2, 3, 4]
    assert to_list(ak._do.local_index(array, axis=1)) == [
        [0, 1, 2],
        [],
        None,
        None,
        [0, 1],
    ]
    assert to_list(ak._do.local_index(array, axis=-2)) == [
        [0, 1, 2],
        [],
        None,
        None,
        [0, 1],
    ]
    assert to_list(ak._do.local_index(array, axis=2)) == [
        [[0, 1, 2], [], [0, 1]],
        [],
        None,
        None,
        [[], [0, 1, 2]],
    ]
    assert to_list(ak._do.local_index(array, axis=-1)) == [
        [[0, 1, 2], [], [0, 1]],
        [],
        None,
        None,
        [[], [0, 1, 2]],
    ]


def test_ByteMaskedArray_combinations():
    content = ak.operations.from_iter(
        [[[0, 1, 2], [], [3, 4]], [], [[5]], [[6, 7, 8, 9]], [[], [10, 11, 12]]],
        highlevel=False,
    )
    mask = ak.index.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    array = ak.highlevel.Array(
        ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    )
    assert to_list(array) == [
        [[0, 1, 2], [], [3, 4]],
        [],
        None,
        None,
        [[], [10, 11, 12]],
    ]
    assert to_list(ak.operations.combinations(array, 2, axis=0)) == [
        ([[0, 1, 2], [], [3, 4]], []),
        ([[0, 1, 2], [], [3, 4]], None),
        ([[0, 1, 2], [], [3, 4]], None),
        ([[0, 1, 2], [], [3, 4]], [[], [10, 11, 12]]),
        ([], None),
        ([], None),
        ([], [[], [10, 11, 12]]),
        (None, None),
        (None, [[], [10, 11, 12]]),
        (None, [[], [10, 11, 12]]),
    ]
    assert to_list(ak.operations.combinations(array, 2, axis=-3)) == [
        ([[0, 1, 2], [], [3, 4]], []),
        ([[0, 1, 2], [], [3, 4]], None),
        ([[0, 1, 2], [], [3, 4]], None),
        ([[0, 1, 2], [], [3, 4]], [[], [10, 11, 12]]),
        ([], None),
        ([], None),
        ([], [[], [10, 11, 12]]),
        (None, None),
        (None, [[], [10, 11, 12]]),
        (None, [[], [10, 11, 12]]),
    ]
    assert to_list(ak.operations.combinations(array, 2, axis=1)) == [
        [([0, 1, 2], []), ([0, 1, 2], [3, 4]), ([], [3, 4])],
        [],
        None,
        None,
        [([], [10, 11, 12])],
    ]
    assert to_list(ak.operations.combinations(array, 2, axis=-2)) == [
        [([0, 1, 2], []), ([0, 1, 2], [3, 4]), ([], [3, 4])],
        [],
        None,
        None,
        [([], [10, 11, 12])],
    ]
    assert to_list(ak.operations.combinations(array, 2, axis=2)) == [
        [[(0, 1), (0, 2), (1, 2)], [], [(3, 4)]],
        [],
        None,
        None,
        [[], [(10, 11), (10, 12), (11, 12)]],
    ]
    assert to_list(ak.operations.combinations(array, 2, axis=-1)) == [
        [[(0, 1), (0, 2), (1, 2)], [], [(3, 4)]],
        [],
        None,
        None,
        [[], [(10, 11), (10, 12), (11, 12)]],
    ]


def test_IndexedOptionArray_combinations():
    content = ak.operations.from_iter(
        [[[0, 1, 2], [], [3, 4]], [], [[5]], [[6, 7, 8, 9]], [[], [10, 11, 12]]],
        highlevel=False,
    )
    index = ak.index.Index64(np.array([0, 1, -1, -1, 4], dtype=np.int64))
    array = ak.Array(ak.contents.IndexedOptionArray(index, content))
    assert to_list(array) == [
        [[0, 1, 2], [], [3, 4]],
        [],
        None,
        None,
        [[], [10, 11, 12]],
    ]
    assert to_list(ak.operations.combinations(array, 2, axis=0)) == [
        ([[0, 1, 2], [], [3, 4]], []),
        ([[0, 1, 2], [], [3, 4]], None),
        ([[0, 1, 2], [], [3, 4]], None),
        ([[0, 1, 2], [], [3, 4]], [[], [10, 11, 12]]),
        ([], None),
        ([], None),
        ([], [[], [10, 11, 12]]),
        (None, None),
        (None, [[], [10, 11, 12]]),
        (None, [[], [10, 11, 12]]),
    ]
    assert to_list(ak.operations.combinations(array, 2, axis=-3)) == [
        ([[0, 1, 2], [], [3, 4]], []),
        ([[0, 1, 2], [], [3, 4]], None),
        ([[0, 1, 2], [], [3, 4]], None),
        ([[0, 1, 2], [], [3, 4]], [[], [10, 11, 12]]),
        ([], None),
        ([], None),
        ([], [[], [10, 11, 12]]),
        (None, None),
        (None, [[], [10, 11, 12]]),
        (None, [[], [10, 11, 12]]),
    ]
    assert to_list(ak.operations.combinations(array, 2, axis=1)) == [
        [([0, 1, 2], []), ([0, 1, 2], [3, 4]), ([], [3, 4])],
        [],
        None,
        None,
        [([], [10, 11, 12])],
    ]
    assert to_list(ak.operations.combinations(array, 2, axis=-2)) == [
        [([0, 1, 2], []), ([0, 1, 2], [3, 4]), ([], [3, 4])],
        [],
        None,
        None,
        [([], [10, 11, 12])],
    ]
    assert to_list(ak.operations.combinations(array, 2, axis=2)) == [
        [[(0, 1), (0, 2), (1, 2)], [], [(3, 4)]],
        [],
        None,
        None,
        [[], [(10, 11), (10, 12), (11, 12)]],
    ]
    assert to_list(ak.operations.combinations(array, 2, axis=-1)) == [
        [[(0, 1), (0, 2), (1, 2)], [], [(3, 4)]],
        [],
        None,
        None,
        [[], [(10, 11), (10, 12), (11, 12)]],
    ]


def test_merge():
    content = ak.operations.from_iter(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    mask = ak.index.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    array1 = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    assert to_list(array1) == [[0.0, 1.1, 2.2], [], None, None, [6.6, 7.7, 8.8, 9.9]]
    array2 = ak.highlevel.Array([[0.0, 1.1, 2.2], [], None, None, [6.6, 7.7, 8.8, 9.9]])
    array12 = ak.operations.concatenate([array1, array2], highlevel=False)
    assert to_list(array12) == [
        [0.0, 1.1, 2.2],
        [],
        None,
        None,
        [6.6, 7.7, 8.8, 9.9],
        [0.0, 1.1, 2.2],
        [],
        None,
        None,
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert isinstance(array12, ak.contents.IndexedOptionArray)
    assert isinstance(
        array12.content, (ak.contents.ListArray, ak.contents.ListOffsetArray)
    )
    assert isinstance(array12.content.content, ak.contents.NumpyArray)
    assert to_list(array12.content.content) == [
        0.0,
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
        6.6,
        7.7,
        8.8,
        9.9,
        0.0,
        1.1,
        2.2,
        6.6,
        7.7,
        8.8,
        9.9,
    ]


def test_BitMaskedArray():
    content = ak.contents.NumpyArray(np.arange(13))
    mask = ak.index.IndexU8(np.array([58, 59], dtype=np.uint8))
    array = ak.contents.BitMaskedArray(
        mask, content, valid_when=False, length=13, lsb_order=False
    )
    assert np.asarray(array.to_ByteMaskedArray().mask).tolist() == [
        0,
        0,
        1,
        1,
        1,
        0,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
    ]
    assert np.asarray(array.to_IndexedOptionArray64().index).tolist() == [
        0,
        1,
        -1,
        -1,
        -1,
        5,
        -1,
        7,
        8,
        9,
        -1,
        -1,
        -1,
    ]
    assert to_list(array) == [
        0,
        1,
        None,
        None,
        None,
        5,
        None,
        7,
        8,
        9,
        None,
        None,
        None,
    ]
    assert (
        ak.operations.to_json(array)
        == "[0,1,null,null,null,5,null,7,8,9,null,null,null]"
    )
    assert to_list(array[1:-1]) == [
        1,
        None,
        None,
        None,
        5,
        None,
        7,
        8,
        9,
        None,
        None,
    ]
    assert to_list(array[8:]) == [8, 9, None, None, None]

    array = ak.contents.BitMaskedArray(
        mask, content, valid_when=False, length=13, lsb_order=True
    )
    assert np.asarray(array.to_ByteMaskedArray().mask).tolist() == [
        0,
        1,
        0,
        1,
        1,
        1,
        0,
        0,
        1,
        1,
        0,
        1,
        1,
    ]
    assert np.asarray(array.to_IndexedOptionArray64().index).tolist() == [
        0,
        -1,
        2,
        -1,
        -1,
        -1,
        6,
        7,
        -1,
        -1,
        10,
        -1,
        -1,
    ]
    assert to_list(array) == [
        0,
        None,
        2,
        None,
        None,
        None,
        6,
        7,
        None,
        None,
        10,
        None,
        None,
    ]
    assert (
        ak.operations.to_json(array)
        == "[0,null,2,null,null,null,6,7,null,null,10,null,null]"
    )
    assert to_list(array[1:-1]) == [
        None,
        2,
        None,
        None,
        None,
        6,
        7,
        None,
        None,
        10,
        None,
    ]
    assert to_list(array[8:]) == [None, None, 10, None, None]


def test_UnmaskedArray():
    content = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5], dtype=np.float64)
    )
    array = ak.contents.UnmaskedArray(content)
    assert to_list(array) == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert str(ak.operations.type(content)) == "5 * float64"
    assert str(ak.operations.type(ak.highlevel.Array(content))) == "5 * float64"
    assert str(ak.operations.type(array)) == "5 * ?float64"
    assert str(ak.operations.type(ak.highlevel.Array(array))) == "5 * ?float64"
    assert str(content.form.type) == "float64"
    assert str(array.form.type) == "?float64"


def test_tomask():
    array = ak.highlevel.Array(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]]
    )
    mask1 = ak.highlevel.Array([True, True, False, False, True])
    assert to_list(array[mask1]) == [[0.0, 1.1, 2.2], [], [6.6, 7.7, 8.8, 9.9]]
    assert to_list(ak.operations.mask(array, mask1)) == [
        [0.0, 1.1, 2.2],
        [],
        None,
        None,
        [6.6, 7.7, 8.8, 9.9],
    ]

    mask2 = ak.highlevel.Array(
        [[False, True, False], [], [True, True], [False], [True, False, False, True]]
    )
    assert to_list(array[mask2]) == [[1.1], [], [3.3, 4.4], [], [6.6, 9.9]]
    assert to_list(ak.operations.mask(array, mask2)) == [
        [None, 1.1, None],
        [],
        [3.3, 4.4],
        [None],
        [6.6, None, None, 9.9],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import cupy as cp  # noqa: F401
import numpy as np
import pytest

import awkward as ak


@pytest.mark.xfail(reason="unimplemented CUDA Kernels (awkward_ByteMaskedArray_numnull")
def test_num_1():
    content = ak.Array(
        ["one", "two", "three", "four", "five", "six", "seven", "eight", "nine"]
    ).layout
    bitmask = ak.index.IndexU8(np.array([40, 34], dtype=np.uint8))
    array = ak.Array(
        ak.contents.bitmaskedarray.BitMaskedArray(bitmask, content, False, 9, False)
    )
    cuda_array = ak.to_backend(array, "cuda")
    assert ak.num(cuda_array, 0) == ak.num(array, 0)
    with pytest.raises(NotImplementedError):
        ak.num(cuda_array, 1)


@pytest.mark.xfail(reason="unimplemented CUDA Kernels (awkward_ByteMaskedArray_numnull")
def test_num_2():
    content = ak.Array(
        ["one", "two", "three", "four", "five", "six", "seven", "eight", "nine"]
    ).layout
    bytemask = ak.index.Index8(np.array([False, True, False], dtype=bool))
    array = ak.Array(
        ak.contents.bytemaskedarray.ByteMaskedArray(bytemask, content, True)
    )
    cuda_array = ak.to_backend(array, "cuda")
    assert ak.num(cuda_array, 0) == ak.num(array, 0)
    with pytest.raises(NotImplementedError):
        ak.num(cuda_array, 1)


def test_num_3():
    array = ak.Array(
        ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5]))
    )
    cuda_array = ak.to_backend(array, "cuda")
    assert ak.num(cuda_array, 0) == ak.num(array, 0)


def test_num_4():
    array = ak.Array(
        ak.contents.numpyarray.NumpyArray(
            np.array([[0.0, 1.1], [2.2, 3.3], [4.4, 5.5]])
        )
    )
    cuda_array = ak.to_backend(array, "cuda")
    assert ak.num(cuda_array, 0) == ak.num(array, 0)
    assert ak.num(cuda_array, 1).tolist() == ak.num(array, 1).tolist()


def test_num_5():
    array = ak.Array(ak.contents.EmptyArray())
    cuda_array = ak.to_backend(array, "cuda")
    assert ak.num(cuda_array, 0) == ak.num(array, 0)


def test_num_6():
    content = ak.contents.numpyarray.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6, 9]))
    array = ak.Array(ak.contents.listoffsetarray.ListOffsetArray(offsets, content))
    cuda_array = ak.to_backend(array, "cuda")
    assert ak.num(cuda_array, 0) == ak.num(array, 0)
    assert ak.num(cuda_array, 1).tolist() == ak.num(array, 1).tolist()


def test_num_7():
    content = ak.contents.numpyarray.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.index.IndexU32(np.array([0, 3, 3, 5, 6, 9]))
    array = ak.Array(ak.contents.listoffsetarray.ListOffsetArray(offsets, content))
    cuda_array = ak.to_backend(array, "cuda")
    assert ak.num(cuda_array, 0) == ak.num(array, 0)
    assert ak.num(cuda_array, 1).tolist() == ak.num(array, 1).tolist()


def test_num_8():
    content = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.10])
    )
    offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.contents.listoffsetarray.ListOffsetArray(offsets, content)
    regulararray = ak.contents.regulararray.RegularArray(listoffsetarray, 2)
    starts = ak.index.Index64(np.array([0, 1]))
    stops = ak.index.Index64(np.array([2, 3]))
    listarray = ak.contents.listarray.ListArray(starts, stops, regulararray)

    cuda_listoffsetarray = ak.to_backend(listoffsetarray, "cuda")
    assert ak.num(cuda_listoffsetarray, 0) == ak.num(ak.Array(listoffsetarray), 0)
    assert (
        ak.num(cuda_listoffsetarray, 1).tolist()
        == ak.num(ak.Array(listoffsetarray), 1).tolist()
    )

    cuda_regulararray = ak.to_backend(regulararray, "cuda")
    assert ak.num(cuda_regulararray, 0) == ak.num(ak.Array(regulararray), 0)
    assert (
        ak.num(cuda_regulararray, 1).tolist()
        == ak.num(ak.Array(regulararray), 1).tolist()
    )

    cuda_listarray = ak.to_backend(listarray, "cuda")
    assert ak.num(cuda_listarray, 0) == ak.num(ak.Array(listarray), 0)
    assert ak.num(cuda_listarray, 1).tolist() == ak.num(ak.Array(listarray), 1).tolist()

    content1 = ak.contents.numpyarray.NumpyArray(np.array([1, 2, 3, 4, 5]))
    content2 = ak.contents.numpyarray.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.index.Index32(np.array([0, 3, 3, 5, 6, 9]))

    recordarray = ak.Array(
        ak.contents.recordarray.RecordArray(
            [content1, listoffsetarray, content2, content1],
            fields=["one", "two", "2", "wonky"],
        )
    )

    cuda_recordarray = ak.to_backend(recordarray, "cuda")

    assert ak.num(cuda_recordarray, 0) == ak.num(recordarray, 0)

    content0 = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]]).layout
    content = ak.Array(
        ["one", "two", "three", "four", "five", "six", "seven", "eight", "nine"]
    ).layout
    tags = ak.index.Index8(np.array([1, 1, 0, 0, 1, 0, 1, 1], dtype=np.int8))
    index = ak.index.Index32(np.array([0, 1, 0, 1, 2, 2, 4, 3], dtype=np.int32))
    unionarray = ak.Array(
        ak.contents.unionarray.UnionArray(tags, index, [content0, content1])
    )

    cuda_unionarray = ak.to_backend(unionarray, "cuda")
    assert ak.num(cuda_unionarray, 0) == ak.num(unionarray, 0)


def test_num_9():
    content = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    index = ak.index.Index32(np.array([0, 2, 4, 6, 8, 9, 7, 5], dtype=np.int64))
    indexedarray = ak.Array(ak.contents.indexedarray.IndexedArray(index, content))

    cuda_indexedarray = ak.to_backend(indexedarray, "cuda")
    assert ak.num(cuda_indexedarray, 0) == ak.num(indexedarray, 0)

    ioa = ak.Array(
        ak.contents.indexedoptionarray.IndexedOptionArray(
            ak.index.Index32([-30, 19, 6, 7, -3, 21, 13, 22, 17, 9, -12, 16]),
            ak.contents.numpyarray.NumpyArray(
                np.array(
                    [
                        5.2,
                        1.7,
                        6.7,
                        -0.4,
                        4.0,
                        7.8,
                        3.8,
                        6.8,
                        4.2,
                        0.3,
                        4.6,
                        6.2,
                        6.9,
                        -0.7,
                        3.9,
                        1.6,
                        8.7,
                        -0.7,
                        3.2,
                        4.3,
                        4.0,
                        5.8,
                        4.2,
                        7.0,
                        5.6,
                        3.8,
                    ]
                )
            ),
        )
    )
    cuda_ioa = ak.to_backend(ioa, "cuda")
    ak.to_backend(cuda_ioa, "cpu")
    assert ak.num(cuda_ioa, 0) == ak.num(ioa, 0)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test_byte_masked_array():
    layout = ak.contents.ByteMaskedArray.simplified(
        ak.index.Index8(np.r_[0, 1, 1]),
        ak.contents.IndexedOptionArray(
            ak.index.Index64([0, 1, -1]), ak.contents.NumpyArray(np.arange(3))
        ),
        valid_when=True,
    )
    assert layout.to_list() == [None, 1, None]

    assert isinstance(layout, ak.contents.IndexedOptionArray)
    assert np.asarray(layout.index).tolist() == [-1, 1, -1]


def test_bit_masked_array():
    layout = ak.contents.BitMaskedArray.simplified(
        ak.index.IndexU8(np.array([0 | 1 << 1 | 1 << 2], dtype=np.uint8)),
        ak.contents.IndexedOptionArray(
            ak.index.Index64([0, 1, -1]), ak.contents.NumpyArray(np.arange(3))
        ),
        valid_when=True,
        length=3,
        lsb_order=True,
    )
    assert layout.to_list() == [None, 1, None]

    assert isinstance(layout, ak.contents.IndexedOptionArray)
    assert np.asarray(layout.index).tolist() == [-1, 1, -1]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("flatten",)
import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import maybe_posaxis, wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._regularize import regularize_axis

np = NumpyMetadata.instance()


@high_level_function
def flatten(array, axis=1, *, highlevel=True, behavior=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        axis (None or int): If None, the operation flattens all levels of
            nesting, returning a 1-dimensional array. Otherwise, it flattens
            at a specified depth. The outermost dimension is `0`, followed
            by `1`, etc., and negative values count backward from the
            innermost: `-1` is the innermost dimension, `-2` is the next
            level up, etc.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Returns an array with one level of nesting removed by erasing the
    boundaries between consecutive lists. Since this operates on a level of
    nesting, `axis=0` is a special case that only removes values at the
    top level that are equal to None.

    Consider the following.

        >>> array = ak.Array([[[1.1, 2.2, 3.3],
        ...                    [],
        ...                    [4.4, 5.5],
        ...                    [6.6]],
        ...                   [],
        ...                   [[7.7],
        ...                    [8.8, 9.9]
        ...                   ]])

    At `axis=1`, the outer lists (length 4, length 0, length 2) become a single
    list (of length 6).

        >>> ak.flatten(array, axis=1).show()
        [[1.1, 2.2, 3.3],
         [],
         [4.4, 5.5],
         [6.6],
         [7.7],
         [8.8, 9.9]]

    At `axis=2`, the inner lists (lengths 3, 0, 2, 1, 1, and 2) become three
    lists (of lengths 6, 0, and 3).

        >>> ak.flatten(array, axis=2).show()
        [[1.1, 2.2, 3.3, 4.4, 5.5, 6.6],
         [],
         [7.7, 8.8, 9.9]]

    There's also an option to completely flatten the array with `axis=None`.
    This is useful for passing the data to a function that doesn't care about
    nested structure, such as a plotting routine.

        >>> ak.flatten(array, axis=None).show()
        [1.1,
         2.2,
         3.3,
         4.4,
         5.5,
         6.6,
         7.7,
         8.8,
         9.9]

    Missing values are eliminated by flattening: there is no distinction
    between an empty list and a value of None at the level of flattening.

        >>> array = ak.Array([[1.1, 2.2, 3.3], None, [4.4], [], [5.5]])
        >>> ak.flatten(array, axis=1)
        <Array [1.1, 2.2, 3.3, 4.4, 5.5] type='5 * float64'>

    As a consequence, flattening at `axis=0` does only one thing: it removes
    None values from the top level.

        >>> ak.flatten(array, axis=0)
        <Array [[1.1, 2.2, 3.3], [4.4], [], [5.5]] type='4 * var * float64'>

    As a technical detail, the flattening operation can be trivial in a common
    case, #ak.contents.ListOffsetArray in which the first `offset` is `0`.
    In that case, the flattened data is simply the array node's `content`.

        >>> array = ak.Array([[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]])
        >>> array.layout
        <ListOffsetArray len='5'>
            <offsets><Index dtype='int64' len='6'>
                [ 0  3  3  5  6 10]
            </Index></offsets>
            <content><NumpyArray dtype='float64' len='10'>
                [0.  1.1 2.2 3.3 4.4 5.5 6.6 7.7 8.8 9.9]
            </NumpyArray></content>
        </ListOffsetArray>

        >>> ak.flatten(array).layout
        <NumpyArray dtype='float64' len='10'>
            [0.  1.1 2.2 3.3 4.4 5.5 6.6 7.7 8.8 9.9]
        </NumpyArray>

        >>> array.layout.content
        <NumpyArray dtype='float64' len='10'>
            [0.  1.1 2.2 3.3 4.4 5.5 6.6 7.7 8.8 9.9]
        </NumpyArray>

    However, it is important to keep in mind that this is a special case:
    #ak.flatten and `content` are not interchangeable!

        >>> array = ak.Array(
        ...     ak.contents.ListArray(
        ...         ak.index.Index64(np.array([ 9, 100, 5, 8, 1])),
        ...         ak.index.Index64(np.array([12, 100, 7, 9, 5])),
        ...         ak.contents.NumpyArray(
        ...             np.array([999, 6.6, 7.7, 8.8, 9.9, 3.3, 4.4, 999, 5.5, 0., 1.1, 2.2, 999])
        ...         ),
        ...     )
        ... )
        >>> array.show()
        [[0, 1.1, 2.2],
         [],
         [3.3, 4.4],
         [5.5],
         [6.6, 7.7, 8.8, 9.9]]

        >>> ak.flatten(array).show()
        [0,
         1.1,
         2.2,
         3.3,
         4.4,
         5.5,
         6.6,
         7.7,
         8.8,
         9.9]

        >>> ak.Array(array.layout.content).show()
        [999,
         6.6,
         7.7,
         8.8,
         9.9,
         3.3,
         4.4,
         999,
         5.5,
         0,
         1.1,
         2.2,
         999]
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, axis, highlevel, behavior)


def _impl(array, axis, highlevel, behavior):
    axis = regularize_axis(axis)
    layout = ak.operations.to_layout(array, allow_record=False, allow_other=False)

    if axis is None:
        out = ak._do.remove_structure(layout, function_name="ak.flatten")
        assert isinstance(out, tuple) and all(
            isinstance(x, ak.contents.Content) for x in out
        )

        result = ak._do.mergemany(out)

        return wrap_layout(result, behavior, highlevel)

    elif axis == 0 or maybe_posaxis(layout, axis, 1) == 0:

        def apply(layout):
            backend = layout.backend

            if layout.is_unknown:
                return apply(ak.contents.NumpyArray(backend.nplike.asarray([])))

            elif layout.is_indexed:
                return apply(layout.project())

            elif layout.is_union:
                if not any(
                    x.is_option and not isinstance(x, ak.contents.UnmaskedArray)
                    for x in layout.contents
                ):
                    return layout

                tags = backend.index_nplike.asarray(layout.tags)
                index = backend.index_nplike.asarray(
                    backend.nplike.asarray(layout.index), copy=True
                )
                bigmask = backend.index_nplike.empty(len(index), dtype=np.bool_)
                for tag, content in enumerate(layout.contents):
                    if content.is_option and not isinstance(
                        content, ak.contents.UnmaskedArray
                    ):
                        bigmask[:] = False
                        bigmask[tags == tag] = backend.index_nplike.asarray(
                            content.mask_as_bool(valid_when=False)
                        ).view(np.bool_)
                        index[bigmask] = -1

                good = index >= 0
                return ak.contents.UnionArray(
                    ak.index.Index8(tags[good]),
                    ak.index.Index64(index[good]),
                    layout.contents,
                )

            elif layout.is_option:
                return layout.project()

            else:
                return layout

        out = apply(layout)

        return wrap_layout(out, behavior, highlevel, like=array)

    else:
        out = ak._do.flatten(layout, axis)
        return wrap_layout(out, behavior, highlevel, like=array)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak


def test_is_none():
    array = ak.Array(
        [
            None,
            [None],
            [{"x": None, "y": None}],
            [{"x": [None], "y": [None]}],
            [{"x": [1], "y": [[None]]}],
            [{"x": [2], "y": [[1, 2, 3]]}],
        ]
    )

    assert ak.is_none(array, axis=0).tolist() == [
        True,
        False,
        False,
        False,
        False,
        False,
    ]
    assert ak.is_none(array, axis=1).tolist() == [
        None,
        [True],
        [False],
        [False],
        [False],
        [False],
    ]
    assert ak.is_none(array, axis=2).tolist() == [
        None,
        [None],
        [{"x": None, "y": None}],
        [{"x": [True], "y": [True]}],
        [{"x": [False], "y": [False]}],
        [{"x": [False], "y": [False]}],
    ]

    with pytest.raises(np.AxisError):
        ak.is_none(array, axis=3)

    assert ak.is_none(array, axis=-1).tolist() == [
        None,
        [None],
        [{"x": None, "y": None}],
        [{"x": [True], "y": [None]}],
        [{"x": [False], "y": [[True]]}],
        [{"x": [False], "y": [[False, False, False]]}],
    ]
    assert ak.is_none(array, axis=-2).tolist() == [
        None,
        [None],
        [{"x": True, "y": None}],
        [{"x": False, "y": [True]}],
        [{"x": False, "y": [False]}],
        [{"x": False, "y": [False]}],
    ]

    with pytest.raises(np.AxisError):
        ak.is_none(array, axis=-3)


def test_singletons():
    array = ak.Array(
        [
            None,
            [None],
            [{"x": None, "y": None}],
            [{"x": [None], "y": [None]}],
            [{"x": [1], "y": [[None]]}],
            [{"x": [2], "y": [[1, 2, 3]]}],
        ]
    )

    assert ak.singletons(array, axis=0).tolist() == [
        [],
        [[None]],
        [[{"x": None, "y": None}]],
        [[{"x": [None], "y": [None]}]],
        [[{"x": [1], "y": [[None]]}]],
        [[{"x": [2], "y": [[1, 2, 3]]}]],
    ]

    assert ak.singletons(array, axis=1).tolist() == [
        None,
        [[]],
        [[{"x": None, "y": None}]],
        [[{"x": [None], "y": [None]}]],
        [[{"x": [1], "y": [[None]]}]],
        [[{"x": [2], "y": [[1, 2, 3]]}]],
    ]

    assert ak.singletons(array, axis=2).tolist() == [
        None,
        [None],
        [{"x": None, "y": None}],
        [{"x": [[]], "y": [[]]}],
        [{"x": [[1]], "y": [[[None]]]}],
        [{"x": [[2]], "y": [[[1, 2, 3]]]}],
    ]

    with pytest.raises(np.AxisError):
        ak.singletons(array, axis=3)

    assert ak.singletons(array, axis=-1).tolist() == [
        None,
        [None],
        [{"x": None, "y": None}],
        [{"x": [[]], "y": [None]}],
        [{"x": [[1]], "y": [[[]]]}],
        [{"x": [[2]], "y": [[[1], [2], [3]]]}],
    ]

    assert ak.singletons(array, axis=-2).tolist() == [
        None,
        [None],
        [{"x": [], "y": None}],
        [{"x": [[None]], "y": [[]]}],
        [{"x": [[1]], "y": [[[None]]]}],
        [{"x": [[2]], "y": [[[1, 2, 3]]]}],
    ]

    with pytest.raises(np.AxisError):
        ak.singletons(array, axis=-3)


def test_is_none_union():
    left = ak.Array([[[{"x": 1, "y": None}]]])
    right = ak.Array([[[{"a": 1, "b": None}]]])

    array = ak.concatenate([left, right], axis=2)

    assert ak.is_none(array, axis=-1).tolist() == [
        [[{"x": False, "y": True}, {"a": False, "b": True}]]
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list


def test():
    first = ak.operations.from_numpy(np.array([1, 2, 3]))
    deltas = ak.highlevel.Array([[1, 2], [1, 2], [1, 2, 3]])
    assert np.hstack(
        (
            first[:, np.newaxis],
            ak.operations.fill_none(ak.operations.pad_none(deltas, 3, axis=-1), 999),
        )
    ).tolist() == [[1, 1, 2, 999], [2, 1, 2, 999], [3, 1, 2, 3]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

import copy
import json
import math
from collections.abc import MutableMapping, Sequence

import awkward as ak
from awkward._backends.backend import Backend
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.numpylike import ArrayLike, IndexType, NumpyMetadata
from awkward._nplikes.shape import ShapeItem, unknown_length
from awkward._nplikes.typetracer import MaybeNone, TypeTracer
from awkward._parameters import (
    type_parameters_equal,
)
from awkward._regularize import is_integer, is_integer_like
from awkward._slicing import NO_HEAD
from awkward._typing import TYPE_CHECKING, Callable, Final, Self, SupportsIndex, final
from awkward._util import UNSET
from awkward.contents.bytemaskedarray import ByteMaskedArray
from awkward.contents.content import Content
from awkward.forms.bitmaskedform import BitMaskedForm
from awkward.forms.form import Form
from awkward.index import Index

if TYPE_CHECKING:
    from awkward._slicing import SliceItem
    from awkward.contents import IndexedOptionArray

np = NumpyMetadata.instance()
numpy = Numpy.instance()


@final
class BitMaskedArray(Content):
    """
    Like #ak.contents.ByteMaskedArray, BitMaskedArray implements an
    #ak.types.OptionType with two buffers, `mask` and `content`.
    However, the boolean `mask` values are packed into a bitmap.

    BitMaskedArray has an additional parameter, `lsb_order`; if True,
    the position of each bit is in
    [Least-Significant Bit order](https://en.wikipedia.org/wiki/Bit_numbering)
    (LSB):

        is_valid[j] = bool(mask[j // 8] & (1 << (j % 8))) == valid_when

    If False, the position of each bit is in Most-Significant Bit order
    (MSB):

        is_valid[j] = bool(mask[j // 8] & (128 >> (j % 8))) == valid_when

    If the logical size of the buffer is not a multiple of 8, the `mask`
    has to be padded. Thus, an explicit `length` is also part of the
    class's definition.

    This is equivalent to *all* of Apache Arrow's array types because they all
    [use bitmaps](https://arrow.apache.org/docs/format/Columnar.html#validity-bitmaps)
    to mask their data, with `valid_when=True` and `lsb_order=True`.

    To illustrate how the constructor arguments are interpreted, the following is a
    simplified implementation of `__init__`, `__len__`, and `__getitem__`:

        class BitMaskedArray(Content):
            def __init__(self, mask, content, valid_when, length, lsb_order):
                assert isinstance(mask, IndexU8)
                assert isinstance(content, Content)
                assert isinstance(valid_when, bool)
                assert isinstance(length, int) and length >= 0
                assert isinstance(lsb_order, bool)
                assert len(mask) <= len(content)
                self.mask = mask
                self.content = content
                self.valid_when = valid_when
                self.length = length
                self.lsb_order = lsb_order

            def __len__(self):
                return self.length

            def __getitem__(self, where):
                if isinstance(where, int):
                    if where < 0:
                        where += len(self)
                    assert 0 <= where < len(self)
                    if self.lsb_order:
                        bit = bool(self.mask[where // 8] & (1 << (where % 8)))
                    else:
                        bit = bool(self.mask[where // 8] & (128 >> (where % 8)))
                    if bit == self.valid_when:
                        return self.content[where]
                    else:
                        return None

                elif isinstance(where, slice) and where.step is None:
                    # In general, slices must convert BitMaskedArray to ByteMaskedArray.
                    bytemask = np.unpackbits(
                        self.mask, bitorder=("little" if self.lsb_order else "big")
                    ).view(bool)
                    return ByteMaskedArray(
                        bytemask[where.start : where.stop],
                        self.content[where.start : where.stop],
                        valid_when=self.valid_when,
                    )

                elif isinstance(where, str):
                    return BitMaskedArray(
                        self.mask,
                        self.content[where],
                        valid_when=self.valid_when,
                        length=self.length,
                        lsb_order=self.lsb_order,
                    )

                else:
                    raise AssertionError(where)
    """

    is_option = True

    def __init__(
        self, mask, content, valid_when, length, lsb_order, *, parameters=None
    ):
        if not (isinstance(mask, Index) and mask.dtype == np.dtype(np.uint8)):
            raise TypeError(
                "{} 'mask' must be an Index with dtype=uint8, not {}".format(
                    type(self).__name__, repr(mask)
                )
            )
        if not isinstance(content, Content):
            raise TypeError(
                "{} 'content' must be a Content subtype, not {}".format(
                    type(self).__name__, repr(content)
                )
            )
        if content.is_union or content.is_indexed or content.is_option:
            raise TypeError(
                "{0} cannot contain a union-type, option-type, or indexed 'content' ({1}); try {0}.simplified instead".format(
                    type(self).__name__, type(content).__name__
                )
            )
        if not isinstance(valid_when, bool):
            raise TypeError(
                "{} 'valid_when' must be boolean, not {}".format(
                    type(self).__name__, repr(valid_when)
                )
            )
        if length is not unknown_length:
            if not (is_integer(length) and length >= 0):
                raise TypeError(
                    "{} 'length' must be a non-negative integer, not {}".format(
                        type(self).__name__, length
                    )
                )
        if not isinstance(lsb_order, bool):
            raise TypeError(
                "{} 'lsb_order' must be boolean, not {}".format(
                    type(self).__name__, repr(lsb_order)
                )
            )
        if (
            not (length is unknown_length or mask.length is unknown_length)
            and length > mask.length * 8
        ):
            raise ValueError(
                "{} 'length' ({}) must be <= len(mask) * 8 ({})".format(
                    type(self).__name__, length, mask.length * 8
                )
            )
        if (
            not (length is unknown_length or content.length is unknown_length)
            and length > content.length * 8
        ):
            raise ValueError(
                "{} 'length' ({}) must be <= len(content) ({})".format(
                    type(self).__name__, length, content.length
                )
            )

        assert mask.nplike is content.backend.index_nplike

        self._mask = mask
        self._content = content
        self._valid_when = valid_when
        self._length = length
        self._lsb_order = lsb_order
        self._init(parameters, content.backend)

    @property
    def mask(self):
        return self._mask

    @property
    def content(self) -> Content:
        return self._content

    @property
    def valid_when(self):
        return self._valid_when

    @property
    def lsb_order(self):
        return self._lsb_order

    form_cls: Final = BitMaskedForm

    def copy(
        self,
        mask=UNSET,
        content=UNSET,
        valid_when=UNSET,
        length=UNSET,
        lsb_order=UNSET,
        *,
        parameters=UNSET,
    ):
        return BitMaskedArray(
            self._mask if mask is UNSET else mask,
            self._content if content is UNSET else content,
            self._valid_when if valid_when is UNSET else valid_when,
            self._length if length is UNSET else length,
            self._lsb_order if lsb_order is UNSET else lsb_order,
            parameters=self._parameters if parameters is UNSET else parameters,
        )

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy(
            mask=copy.deepcopy(self._mask, memo),
            content=copy.deepcopy(self._content, memo),
            parameters=copy.deepcopy(self._parameters, memo),
        )

    @classmethod
    def simplified(
        cls,
        mask,
        content,
        valid_when,
        length,
        lsb_order,
        *,
        parameters=None,
    ):
        if content.is_union or content.is_indexed or content.is_option:
            backend = content.backend
            index = ak.index.Index64.empty(mask.length * 8, backend.index_nplike)
            backend.maybe_kernel_error(
                backend[
                    "awkward_BitMaskedArray_to_IndexedOptionArray",
                    index.dtype.type,
                    mask.dtype.type,
                ](
                    index.raw(backend.nplike),
                    mask.data,
                    mask.length,
                    valid_when,
                    lsb_order,
                ),
            )
            if content.is_union:
                return content._union_of_optionarrays(index[0:length], parameters)
            else:
                return ak.contents.IndexedOptionArray.simplified(
                    index[0:length], content, parameters=parameters
                )
        else:
            return cls(
                mask,
                content,
                valid_when,
                length,
                lsb_order,
                parameters=parameters,
            )

    def _form_with_key(self, getkey: Callable[[Content], str | None]) -> BitMaskedForm:
        form_key = getkey(self)
        return self.form_cls(
            self._mask.form,
            self._content._form_with_key(getkey),
            self._valid_when,
            self._lsb_order,
            parameters=self._parameters,
            form_key=form_key,
        )

    def _to_buffers(
        self,
        form: Form,
        getkey: Callable[[Content, Form, str], str],
        container: MutableMapping[str, ArrayLike],
        backend: Backend,
        byteorder: str,
    ):
        assert isinstance(form, self.form_cls)
        key = getkey(self, form, "mask")
        container[key] = ak._util.native_to_byteorder(
            self._mask.raw(backend.index_nplike), byteorder
        )
        self._content._to_buffers(form.content, getkey, container, backend, byteorder)

    def _to_typetracer(self, forget_length: bool) -> Self:
        tt = TypeTracer.instance()
        return BitMaskedArray(
            self._mask.to_nplike(tt),
            self._content._to_typetracer(False),
            self._valid_when,
            unknown_length if forget_length else self.length,
            self._lsb_order,
            parameters=self._parameters,
        )

    def _touch_data(self, recursive: bool):
        self._mask._touch_data()
        if recursive:
            self._content._touch_data(recursive)

    def _touch_shape(self, recursive: bool):
        self._mask._touch_shape()
        if recursive:
            self._content._touch_shape(recursive)

    @property
    def length(self) -> ShapeItem:
        return self._length

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<BitMaskedArray valid_when="]
        out.append(repr(json.dumps(self._valid_when)))
        out.append(" lsb_order=")
        out.append(repr(json.dumps(self._lsb_order)))
        out.append(" len=")
        out.append(repr(str(self.length)))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")
        out.append(self._mask._repr(indent + "    ", "<mask>", "</mask>\n"))
        out.append(self._content._repr(indent + "    ", "<content>", "</content>\n"))
        out.append(indent + "</BitMaskedArray>")
        out.append(post)
        return "".join(out)

    def to_IndexedOptionArray64(self) -> IndexedOptionArray:
        index = ak.index.Index64.empty(
            self._mask.length * 8, self._backend.index_nplike
        )
        assert (
            index.nplike is self._backend.nplike
            and self._mask.nplike is self._backend.nplike
        )
        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_BitMaskedArray_to_IndexedOptionArray",
                index.dtype.type,
                self._mask.dtype.type,
            ](
                index.raw(self._backend.nplike),
                self._mask.data,
                self._mask.length,
                self._valid_when,
                self._lsb_order,
            ),
        )
        return ak.contents.IndexedOptionArray(
            index[0 : self._length], self._content, parameters=self._parameters
        )

    def to_ByteMaskedArray(self):
        bytemask = ak.index.Index8.empty(
            self._mask.length * 8, self._backend.index_nplike
        )
        assert (
            bytemask.nplike is self._backend.nplike
            and self._mask.nplike is self._backend.nplike
        )
        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_BitMaskedArray_to_ByteMaskedArray",
                bytemask.dtype.type,
                self._mask.dtype.type,
            ](
                bytemask.data,
                self._mask.data,
                self._mask.length,
                False,  # this differs from the kernel call in 'bytemask'
                self._lsb_order,
            )
        )
        return ByteMaskedArray(
            bytemask[: self._length],
            self._content,
            self._valid_when,
            parameters=self._parameters,
        )

    def to_BitMaskedArray(self, valid_when, lsb_order):
        if lsb_order == self._lsb_order:
            if valid_when == self._valid_when:
                return self
            else:
                return BitMaskedArray(
                    ak.index.IndexU8(
                        self._backend.index_nplike.logical_not(self._mask.data)
                    ),
                    self._content,
                    valid_when,
                    self._length,
                    lsb_order,
                    parameters=self._parameters,
                )

        else:
            bit_order = "little" if lsb_order else "big"
            bytemask = self.backend.index_nplike.unpackbits(
                self._mask.data, bitorder=bit_order
            )
            bitmask = self.backend.index_nplike.packbits(bytemask, bitorder=bit_order)

            if valid_when != self._valid_when:
                bitmask = ~bitmask

            return BitMaskedArray(
                ak.index.IndexU8(bitmask),
                self._content,
                valid_when,
                self._length,
                lsb_order,
                parameters=self._parameters,
            )

    def mask_as_bool(self, valid_when=None):
        if valid_when is None:
            valid_when = self._valid_when

        bytemask = ak.index.Index8.empty(
            self._mask.length * 8, self._backend.index_nplike
        )
        assert (
            bytemask.nplike is self._backend.nplike
            and self._mask.nplike is self._backend.nplike
        )
        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_BitMaskedArray_to_ByteMaskedArray",
                bytemask.dtype.type,
                self._mask.dtype.type,
            ](
                bytemask.data,
                self._mask.data,
                self._mask.length,
                0 if valid_when == self._valid_when else 1,
                self._lsb_order,
            )
        )
        return bytemask.data[: self._length].view(np.bool_)

    def _getitem_nothing(self):
        return self._content._getitem_range(0, 0)

    def _getitem_at(self, where: IndexType):
        if not self._backend.nplike.known_data:
            self._touch_data(recursive=False)
            return MaybeNone(self._content._getitem_at(where))

        if where < 0:
            where += self.length
        if not (0 <= where < self.length) and self._backend.nplike.known_data:
            raise ak._errors.index_error(self, where)
        if self._lsb_order:
            bit = bool(self._mask[where // 8] & (1 << (where % 8)))
        else:
            bit = bool(self._mask[where // 8] & (128 >> (where % 8)))
        if bit == self._valid_when:
            return self._content._getitem_at(where)
        else:
            return None

    def _getitem_range(self, start: IndexType, stop: IndexType) -> Content:
        return self.to_ByteMaskedArray()._getitem_range(start, stop)

    def _getitem_field(
        self, where: str | SupportsIndex, only_fields: tuple[str, ...] = ()
    ) -> Content:
        return BitMaskedArray.simplified(
            self._mask,
            self._content._getitem_field(where, only_fields),
            self._valid_when,
            self._length,
            self._lsb_order,
            parameters=None,
        )

    def _getitem_fields(
        self, where: list[str | SupportsIndex], only_fields: tuple[str, ...] = ()
    ) -> Content:
        return BitMaskedArray.simplified(
            self._mask,
            self._content._getitem_fields(where, only_fields),
            self._valid_when,
            self._length,
            self._lsb_order,
            parameters=None,
        )

    def _carry(self, carry: Index, allow_lazy: bool) -> Content:
        assert isinstance(carry, ak.index.Index)
        return self.to_ByteMaskedArray()._carry(carry, allow_lazy)

    def _getitem_next_jagged(
        self, slicestarts: Index, slicestops: Index, slicecontent: Content, tail
    ) -> Content:
        return self.to_ByteMaskedArray()._getitem_next_jagged(
            slicestarts, slicestops, slicecontent, tail
        )

    def _getitem_next(
        self,
        head: SliceItem | tuple,
        tail: tuple[SliceItem, ...],
        advanced: Index | None,
    ) -> Content:
        if head is NO_HEAD:
            return self

        elif is_integer_like(head) or isinstance(
            head, (slice, ak.index.Index64, ak.contents.ListOffsetArray)
        ):
            return self.to_ByteMaskedArray()._getitem_next(head, tail, advanced)

        elif isinstance(head, str):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak.contents.IndexedOptionArray):
            return self._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def project(self, mask=None):
        return self.to_ByteMaskedArray().project(mask)

    def _offsets_and_flattened(self, axis: int, depth: int) -> tuple[Index, Content]:
        return self.to_ByteMaskedArray._offsets_and_flattened(axis, depth)

    def _mergeable_next(self, other: Content, mergebool: bool) -> bool:
        # Is the other content is an identity, or a union?
        if other.is_identity_like or other.is_union:
            return True
        # We can only combine option types whose array-record parameters agree
        elif other.is_option or other.is_indexed:
            return self._content._mergeable_next(
                other.content, mergebool
            ) and type_parameters_equal(self._parameters, other._parameters)
        else:
            return self._content._mergeable_next(other, mergebool)

    def _reverse_merge(self, other):
        return self.to_IndexedOptionArray64()._reverse_merge(other)

    def _mergemany(self, others: Sequence[Content]) -> Content:
        if len(others) == 0:
            return self

        out = self.to_IndexedOptionArray64()._mergemany(others)

        if all(
            isinstance(x, BitMaskedArray)
            and x._valid_when == self._valid_when
            and x._lsb_order == self._lsb_order
            for x in others
        ):
            return out.to_BitMaskedArray(self._valid_when, self._lsb_order)
        else:
            return out

    def _fill_none(self, value: Content) -> Content:
        return self.to_IndexedOptionArray64()._fill_none(value)

    def _local_index(self, axis, depth):
        return self.to_ByteMaskedArray()._local_index(axis, depth)

    def _numbers_to_type(self, name, including_unknown):
        return self.to_ByteMaskedArray()._numbers_to_type(name, including_unknown)

    def _is_unique(self, negaxis, starts, parents, outlength):
        if self._mask.length == 0:
            return True
        return self.to_IndexedOptionArray64()._is_unique(
            negaxis, starts, parents, outlength
        )

    def _unique(self, negaxis, starts, parents, outlength):
        if self._mask.length == 0:
            return self
        out = self.to_IndexedOptionArray64()._unique(
            negaxis, starts, parents, outlength
        )
        if negaxis is None:
            return out
        else:
            return out._content

    def _argsort_next(
        self, negaxis, starts, shifts, parents, outlength, ascending, stable
    ):
        return self.to_IndexedOptionArray64()._argsort_next(
            negaxis, starts, shifts, parents, outlength, ascending, stable
        )

    def _sort_next(self, negaxis, starts, parents, outlength, ascending, stable):
        return self.to_IndexedOptionArray64()._sort_next(
            negaxis, starts, parents, outlength, ascending, stable
        )

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        return self.to_ByteMaskedArray()._combinations(
            n, replacement, recordlookup, parameters, axis, depth
        )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
        behavior,
    ):
        return self.to_ByteMaskedArray()._reduce_next(
            reducer,
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            mask,
            keepdims,
            behavior,
        )

    def _validity_error(self, path):
        if self.mask.length * 8 < self.length:
            return f"at {path} ({type(self)!r}): len(mask) * 8 < length"
        elif self._content.length < self.length:
            return f"at {path} ({type(self)!r}): len(content) < length"
        else:
            return self._content._validity_error(path + ".content")

    def _nbytes_part(self):
        return self.mask._nbytes_part() + self.content._nbytes_part()

    def _pad_none(self, target, axis, depth, clip):
        return self.to_ByteMaskedArray()._pad_none(target, axis, depth, clip)

    def _to_arrow(self, pyarrow, mask_node, validbytes, length, options):
        return self.to_ByteMaskedArray()._to_arrow(
            pyarrow, mask_node, validbytes, length, options
        )

    def _to_backend_array(self, allow_missing, backend):
        return self.to_ByteMaskedArray()._to_backend_array(allow_missing, backend)

    def _remove_structure(self, backend, options):
        branch, depth = self.branch_depth
        if branch or options["drop_nones"] or depth > 1:
            return self.project()._remove_structure(backend, options)
        else:
            return [self]

    def _drop_none(self) -> Content:
        return self.to_ByteMaskedArray()._drop_none()

    def _recursively_apply(
        self, action, behavior, depth, depth_context, lateral_context, options
    ):
        if self._backend.nplike.known_data:
            content = self._content[0 : self._length]
        else:
            content = self._content

        if options["return_array"]:
            if options["return_simplified"]:
                make = BitMaskedArray.simplified
            else:
                make = BitMaskedArray

            def continuation():
                return make(
                    self._mask,
                    content._recursively_apply(
                        action,
                        behavior,
                        depth,
                        copy.copy(depth_context),
                        lateral_context,
                        options,
                    ),
                    self._valid_when,
                    self._length,
                    self._lsb_order,
                    parameters=self._parameters if options["keep_parameters"] else None,
                )

        else:

            def continuation():
                content._recursively_apply(
                    action,
                    behavior,
                    depth,
                    copy.copy(depth_context),
                    lateral_context,
                    options,
                )

        result = action(
            self,
            depth=depth,
            depth_context=depth_context,
            lateral_context=lateral_context,
            continuation=continuation,
            behavior=behavior,
            backend=self._backend,
            options=options,
        )

        if isinstance(result, Content):
            return result
        elif result is None:
            return continuation()
        else:
            raise AssertionError(result)

    def to_packed(self) -> Self:
        if self._content.is_record:
            next = self.to_IndexedOptionArray64()

            content = next._content.to_packed()
            if content.length > self._length:
                content = content[: self._length]

            return ak.contents.IndexedOptionArray(
                next._index, content, parameters=next._parameters
            )

        else:
            excess_length = int(math.ceil(self._length / 8.0))
            if self._mask.length == excess_length:
                mask = self._mask
            else:
                mask = self._mask[:excess_length]

            content = self._content.to_packed()
            if content.length > self._length:
                content = content[: self._length]

            return BitMaskedArray(
                mask,
                content,
                self._valid_when,
                self._length,
                self._lsb_order,
                parameters=self._parameters,
            )

    def _to_list(self, behavior, json_conversions):
        if not self._backend.nplike.known_data:
            raise TypeError("cannot convert typetracer arrays to Python lists")

        out = self._to_list_custom(behavior, json_conversions)
        if out is not None:
            return out

        mask = self.mask_as_bool(valid_when=True)[: self._length]
        out = self._content._getitem_range(0, self._length)._to_list(
            behavior, json_conversions
        )

        for i, isvalid in enumerate(mask):
            if not isvalid:
                out[i] = None

        return out

    def _to_backend(self, backend: Backend) -> Self:
        content = self._content.to_backend(backend)
        mask = self._mask.to_nplike(backend.index_nplike)
        return BitMaskedArray(
            mask,
            content,
            valid_when=self._valid_when,
            length=len(self),
            lsb_order=self._lsb_order,
            parameters=self._parameters,
        )

    def _is_equal_to(self, other, index_dtype, numpyarray):
        return (
            self.valid_when == other.valid_when
            and self.lsb_order == other.lsb_order
            and self.mask.is_equal_to(other.mask, index_dtype, numpyarray)
            and self.content.is_equal_to(other.content, index_dtype, numpyarray)
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test_bool():
    result = ak.full_like([True], 2, dtype=np.int64)
    assert ak.almost_equal(result, np.asarray([2], dtype=np.int64))


def test_empty():
    result = ak.full_like([], 2, dtype=np.int64)
    assert result.layout.is_unknown
    result = ak.full_like([], 2, dtype=np.int64, including_unknown=True)
    assert result.layout.is_numpy
    assert result.layout.dtype == np.dtype(np.int64)


def test_complex128():
    result = ak.full_like([1, 2], 4j, dtype=np.complex128)
    assert ak.almost_equal(result, np.asarray([0 + 4j, 0 + 4j], dtype=np.complex128))



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

__all__ = ("almost_equal",)
from awkward._backends.dispatch import backend_of
from awkward._behavior import behavior_of, get_array_class, get_record_class
from awkward._dispatch import high_level_function
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._parameters import parameters_are_equal
from awkward.operations.ak_to_layout import to_layout

np = NumpyMetadata.instance()


@high_level_function
def almost_equal(
    left,
    right,
    *,
    rtol: float = 1e-5,
    atol: float = 1e-8,
    dtype_exact: bool = True,
    check_parameters: bool = True,
    check_regular: bool = True,
):
    """
    Args:
        left: Array-like data (anything #ak.to_layout recognizes).
        right: Array-like data (anything #ak.to_layout recognizes).
        rtol: the relative tolerance parameter (see below).
        atol: the absolute tolerance parameter (see below).
        dtype_exact: whether the dtypes must be exactly the same, or just the
            same family.
        check_parameters: whether to compare parameters.
        check_regular: whether to consider ragged and regular dimensions as
            unequal.

    Return True if the two array-like arguments are considered equal for the
    given options. Otherwise, return False.

    The relative difference (`rtol * abs(b)`) and the absolute difference `atol`
    are added together to compare against the absolute difference between `left`
    and `right`.

    TypeTracer arrays are not supported, as there is very little information to
    be compared.
    """
    # Dispatch
    yield left, right

    # Implementation
    left_behavior = behavior_of(left)
    right_behavior = behavior_of(right)

    left = to_layout(left, allow_record=False).to_packed()
    right = to_layout(right, allow_record=False).to_packed()

    backend = backend_of(left, right)
    if not backend.nplike.known_data:
        raise NotImplementedError(
            "Awkward Arrays with typetracer backends cannot yet be compared with `ak.almost_equal`."
        )

    def is_approx_dtype(left, right) -> bool:
        if not dtype_exact:
            for family in np.integer, np.floating:
                if np.issubdtype(left, family):
                    return np.issubdtype(right, family)
        return left == right

    def packed_list_content(layout):
        layout = layout.to_ListOffsetArray64(False)
        return layout.content[layout.offsets[0] : layout.offsets[-1]]

    def visitor(left, right) -> bool:
        # First, erase indexed types!
        if left.is_indexed and not left.is_option:
            left = left.project()
        if right.is_indexed and not right.is_option:
            right = right.project()

        # Simplify option types
        if left.is_option:
            left = left.to_IndexedOptionArray64()
        if right.is_option:
            right = right.to_IndexedOptionArray64()

        # Simplify regular NumPy types
        if left.is_numpy and left.purelist_depth > 1:
            left = left.to_RegularArray()
        if right.is_numpy and right.purelist_depth > 1:
            right = right.to_RegularArray()

        # Different lengths aren't equal!
        if left.length != right.length:
            return False

        if check_parameters and not parameters_are_equal(
            left.parameters, right.parameters
        ):
            return False

        # Require that the arrays have the same evaluated types
        if not (
            get_array_class(left, left_behavior)
            is get_array_class(right, right_behavior)
            or not check_parameters
        ):
            return False

        # Regular-regular
        if left.is_regular and right.is_regular:
            return (left.size == right.size) and visitor(left.content, right.content)
        # List-list
        elif left.is_list and right.is_list:
            # Mixed regular-var
            if left.is_regular and not right.is_regular:
                return (not check_regular) and visitor(
                    left.content,
                    packed_list_content(right),
                )
            elif right.is_regular and not left.is_regular:
                return (not check_regular) and visitor(
                    packed_list_content(left),
                    right.content,
                )
            else:
                return visitor(
                    packed_list_content(left),
                    packed_list_content(right),
                )

        elif left.is_numpy and right.is_numpy:
            # Timelike types must be exactly compared, including their units
            if (
                np.issubdtype(left.dtype, np.datetime64)
                or np.issubdtype(right.dtype, np.datetime64)
                or np.issubdtype(left.dtype, np.timedelta64)
                or np.issubdtype(right.dtype, np.timedelta64)
            ):
                return (
                    (left.dtype == right.dtype)
                    and backend.nplike.all(left.data == right.data)
                    and left.shape == right.shape
                )
            else:
                return (
                    is_approx_dtype(left.dtype, right.dtype)
                    and backend.nplike.all(
                        backend.nplike.isclose(
                            left.data, right.data, rtol=rtol, atol=atol, equal_nan=False
                        )
                    )
                    and left.shape == right.shape
                )
        elif left.is_option and right.is_option:
            return backend.index_nplike.array_equal(
                left.mask_as_bool(True), right.mask_as_bool(True)
            ) and visitor(left.project(), right.project())
        elif left.is_union and right.is_union:
            # For two unions with different content orderings to match, the tags should be equal at each index
            # Therefore, we can order the contents by index appearance
            def ordered_unique_values(values):
                # First, find unique values and their appearance (from smallest to largest)
                # unique_index is in ascending order of `unique` value
                (
                    unique,
                    unique_index,
                    *_,
                ) = backend.index_nplike.unique_all(values)
                # Now re-order `unique` by order of appearance (`unique_index`)
                return values[backend.index_nplike.sort(unique_index)]

            # Find order of appearance for each union tags, and assume these are one-to-one maps
            left_tag_order = ordered_unique_values(left.tags.data)
            right_tag_order = ordered_unique_values(right.tags.data)

            # Create map from left tags to right tags
            left_tag_to_right_tag = backend.index_nplike.empty(
                left_tag_order.size, dtype=np.int64
            )
            left_tag_to_right_tag[left_tag_order] = right_tag_order

            # Map left tags onto right, such that the result should equal right.tags
            # if the two tag arrays are equivalent
            new_left_tag = left_tag_to_right_tag[left.tags.data]
            if not backend.index_nplike.all(new_left_tag == right.tags.data):
                return False

            # Now project out the contents, and check for equality
            for i, j in zip(left_tag_order, right_tag_order):
                if not visitor(left.project(i), right.project(j)):
                    return False
            return True

        elif left.is_record and right.is_record:
            return (
                (
                    get_record_class(left, left_behavior)
                    is get_record_class(right, right_behavior)
                    or not check_parameters
                )
                and left.is_tuple == right.is_tuple
                and (left.is_tuple or (len(left.fields) == len(right.fields)))
                and all(visitor(left.content(f), right.content(f)) for f in left.fields)
            )
        elif left.is_unknown and right.is_unknown:
            return True

        else:
            return False

    return visitor(left, right)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import cupy as cp
import numpy as np

import awkward as ak


def test_from_cupy():
    cupy_array_1d = cp.arange(10)
    cupy_array_2d = cp.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6], [7.7, 8.8]])

    ak_cupy_array_1d = ak.from_cupy(cupy_array_1d)
    ak_cupy_array_2d = ak.from_cupy(cupy_array_2d)

    for i in range(10):
        assert ak_cupy_array_1d[i] == cupy_array_1d[i]

    for i in range(4):
        for j in range(2):
            assert ak_cupy_array_2d[i][j] == cupy_array_2d[i][j]


def test_from_cupy_tolist():
    cupy_array_1d = cp.array([9, 8, 7, 6, 5, 4, 3, 2, 1, 0])

    ak_cupy_array_1d = ak.from_cupy(cupy_array_1d)

    assert ak.to_list(ak_cupy_array_1d.layout) == [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]


def test_NumpyArray_constructor():
    assert ak.backend(ak.contents.NumpyArray(np.array([1, 2, 3]))) == "cpu"
    assert ak.backend(ak.contents.NumpyArray(cp.array([1, 2, 3]))) == "cuda"


def test_add():
    one = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]], backend="cuda")
    two = ak.Array([100, 200, 300], backend="cuda")
    assert ak.backend(one) == "cuda"
    assert ak.backend(two) == "cuda"
    three = one + two
    assert ak.to_list(three) == [[101.1, 102.2, 103.3], [], [304.4, 305.5]]
    assert ak.backend(three) == "cuda"


def test_add_2():
    one = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]], backend="cuda")
    two = 100
    assert ak.backend(one) == "cuda"
    three = one + two
    assert ak.to_list(three) == [[101.1, 102.2, 103.3], [], [104.4, 105.5]]
    assert ak.backend(three) == "cuda"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak


def test_reducers():
    array = ak.Array([[{"rho": -1.1, "phi": -0.1}, {"rho": 1.1, "phi": 0.1}]])

    with pytest.raises(TypeError):
        assert ak.to_list(ak.operations.all(array, axis=1)) == [
            {"phi": True, "rho": True}
        ]

    with pytest.raises(TypeError):
        assert ak.to_list(ak.operations.any(array, axis=1)) == [
            {"phi": True, "rho": True}
        ]

    with pytest.raises(TypeError):
        assert ak.to_list(ak.operations.argmax(array, axis=1)) == [
            {"phi": True, "rho": True}
        ]

    with pytest.raises(TypeError):
        assert ak.to_list(ak.operations.argmin(array, axis=1)) == [{"phi": 0, "rho": 0}]

    with pytest.raises(TypeError):
        assert ak.to_list(ak.operations.count_nonzero(array, axis=1)) == [
            {"phi": 2, "rho": 2}
        ]

    with pytest.raises(TypeError):
        assert ak.to_list(ak.operations.count(array, axis=1)) == [{"phi": 2, "rho": 2}]

    with pytest.raises(TypeError):
        assert ak.to_list(ak.operations.max(array, axis=1)) == [
            {"phi": 0.1, "rho": 1.1}
        ]

    with pytest.raises(TypeError):
        assert ak.to_list(ak.operations.min(array, axis=1)) == [
            {"phi": -0.1, "rho": -1.1}
        ]

    with pytest.raises(TypeError):
        assert ak.to_list(ak.operations.prod(array, axis=1)) == [
            {"phi": -0.010000000000000002, "rho": -1.2100000000000002}
        ]

    with pytest.raises(TypeError):
        assert ak.to_list(ak.operations.sum(array, axis=1)) == [
            {"phi": 0.0, "rho": 0.0}
        ]


def test_overloaded_reducers():
    def overload_add(array, mask):
        return ak.contents.RecordArray(
            [
                ak.sum(array["rho"], axis=-1, mask_identity=False, highlevel=False),
                ak.sum(array["phi"], axis=-1, mask_identity=False, highlevel=False),
            ],
            ["rho", "phi"],
        )

    behavior = {}
    behavior[ak.sum, "VectorArray2D"] = overload_add

    array = ak.Array(
        [[{"rho": -1.1, "phi": -0.1}, {"rho": 1.1, "phi": 0.1}]],
        with_name="VectorArray2D",
        behavior=behavior,
    )

    assert ak.to_list(ak.sum(array, axis=1)) == [{"rho": 0, "phi": 0}]

    with pytest.raises(
        TypeError, match=r"overloads for custom types: VectorArray2D, int"
    ):
        ak.to_list(array + 1)

    def overload_add2(array, mask):
        return ak.contents.RecordArray(
            [ak.contents.NumpyArray(np.asarray([2.4, 3, 4.5, 6], dtype=np.float64))],
            ["rho"],
        )

    behavior = {}
    behavior[ak.sum, "VectorArray2D"] = overload_add2

    array = ak.Array(
        [[{"rho": -1.1, "phi": -0.1}, {"rho": 1.1, "phi": 0.1}]],
        with_name="VectorArray2D",
        behavior=behavior,
    )

    assert ak.to_list(ak.sum(array, axis=1)) == [{"rho": 2.4}]

    array = ak.highlevel.Array(
        ak.contents.ByteMaskedArray(
            ak.index.Index8(np.array([True, True])),
            ak.contents.RecordArray(
                [
                    ak.contents.NumpyArray(np.arange(2)),
                    ak.contents.IndexedOptionArray(
                        ak.index.Index64(np.array([0, -1])),
                        ak.contents.NumpyArray(np.array([1], dtype=np.int64)),
                    ),
                ],
                ["a", "b"],
            ),
            valid_when=True,
        )
    )

    with pytest.raises(TypeError, match=r"overloads for custom types: a, b"):
        ak.to_list(ak.sum(array, axis=0))

    def overload_argmax(array, mask):
        return ak.argmax(array["rho"], axis=-1, mask_identity=False, highlevel=False)

    behavior = {}
    behavior[ak.argmax, "VectorArray2D"] = overload_argmax

    array = ak.Array(
        [
            [
                {"rho": -1.1, "phi": -0.1},
                {"rho": 1.1, "phi": 0.1},
                {"rho": -1.1, "phi": 0.3},
            ],
            [
                {"rho": 1.1, "phi": 0.1},
                {"rho": -1.1, "phi": 0.3},
                {"rho": -1.1, "phi": -0.1},
            ],
        ],
        with_name="VectorArray2D",
        behavior=behavior,
    )

    assert ak.to_list(ak.argmax(array, axis=1)) == [1, 0]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy
import pytest

import awkward as ak


def test_record():
    ## record → record
    result = ak.enforce_type(
        ak.to_layout([{"x": [1, 2]}], regulararray=False),
        ak.types.from_datashape("{x: var * int64}", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.RecordArray(
            [
                ak.contents.ListOffsetArray(
                    ak.index.Index(numpy.array([0, 2], dtype=numpy.int64)),
                    ak.contents.NumpyArray(numpy.array([1, 2], dtype=numpy.int64)),
                )
            ],
            ["x"],
        )
    )

    ## record → record
    result = ak.enforce_type(
        ak.to_layout([{"x": [1, 0]}], regulararray=False),
        ak.types.from_datashape("{x: var * bool}", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.RecordArray(
            [
                ak.contents.ListOffsetArray(
                    ak.index.Index(numpy.array([0, 2], dtype=numpy.int64)),
                    ak.contents.NumpyArray(numpy.array([1, 0], dtype=numpy.bool_)),
                )
            ],
            ["x"],
        )
    )

    ## record → different tuple
    with pytest.raises(ValueError, match=r"converted between records and tuples"):
        ak.enforce_type(
            ak.to_layout([{"x": [1, 2]}], regulararray=False),
            ak.types.from_datashape("(var * float64)", highlevel=False),
        )

    with pytest.raises(
        TypeError, match=r"can only add new fields to a record if they are option types"
    ):
        ak.enforce_type(
            ak.to_layout([{"x": [1, 2]}], regulararray=False),
            ak.types.from_datashape("{y: var * float64}", highlevel=False),
        )

    ## record → totally different record
    result = ak.enforce_type(
        ak.to_layout([{"x": [1, 2]}], regulararray=False),
        ak.types.from_datashape("{y: ?var * float64}", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.RecordArray(
            [
                ak.contents.IndexedOptionArray(
                    ak.index.Index64([-1]),
                    ak.contents.ListOffsetArray(
                        ak.index.Index(numpy.array([0], dtype=numpy.int64)),
                        ak.contents.NumpyArray(numpy.array([], dtype=numpy.float64)),
                    ),
                )
            ],
            ["y"],
        )
    )
    ## record → extended record
    result = ak.enforce_type(
        ak.to_layout([{"x": [1, 2]}], regulararray=False),
        ak.types.from_datashape("{x: var * int64, y: ?int64}", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.RecordArray(
            [
                ak.contents.ListOffsetArray(
                    ak.index.Index(numpy.array([0, 2], dtype=numpy.int64)),
                    ak.contents.NumpyArray(numpy.array([1, 2], dtype=numpy.int64)),
                ),
                ak.contents.IndexedOptionArray(
                    ak.index.Index64([-1]),
                    ak.contents.NumpyArray(numpy.array([], dtype=numpy.int64)),
                ),
            ],
            ["x", "y"],
        ),
    )
    ## record → empty record
    result = ak.enforce_type(
        ak.to_layout([{"x": [1, 2]}], regulararray=False),
        ak.types.from_datashape("{}", highlevel=False),
    )
    assert result.layout.is_equal_to(ak.contents.RecordArray([], [], length=1))

    ############

    ## tuple → tuple
    result = ak.enforce_type(
        ak.to_layout([([1, 2],)], regulararray=False),
        ak.types.from_datashape("(var * int64)", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.RecordArray(
            [
                ak.contents.ListOffsetArray(
                    ak.index.Index(numpy.array([0, 2], dtype=numpy.int64)),
                    ak.contents.NumpyArray(numpy.array([1, 2], dtype=numpy.int64)),
                )
            ],
            None,
        )
    )

    ## tuple → tuple
    result = ak.enforce_type(
        ak.to_layout([([1, 0],)], regulararray=False),
        ak.types.from_datashape("(var * bool)", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.RecordArray(
            [
                ak.contents.ListOffsetArray(
                    ak.index.Index(numpy.array([0, 2], dtype=numpy.int64)),
                    ak.contents.NumpyArray(numpy.array([1, 0], dtype=numpy.bool_)),
                )
            ],
            None,
        )
    )

    ## tuple → different record
    with pytest.raises(ValueError, match=r"converted between records and tuples"):
        ak.enforce_type(
            ak.to_layout([([1, 2],)], regulararray=False),
            ak.types.from_datashape("{x: var * float64}", highlevel=False),
        )

    with pytest.raises(
        TypeError, match=r"can only add new slots to a tuple if they are option types"
    ):
        ak.enforce_type(
            ak.to_layout([([1, 2],)], regulararray=False),
            ak.types.from_datashape("(var * int64, float32)", highlevel=False),
        )

    ## tuple → extended tuple
    result = ak.enforce_type(
        ak.to_layout([([1, 2],)], regulararray=False),
        ak.types.from_datashape("(var * int64, ?float32)", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.RecordArray(
            [
                ak.contents.ListOffsetArray(
                    ak.index.Index(numpy.array([0, 2], dtype=numpy.int64)),
                    ak.contents.NumpyArray(numpy.array([1, 2], dtype=numpy.int64)),
                ),
                ak.contents.IndexedOptionArray(
                    ak.index.Index64([-1]),
                    ak.contents.NumpyArray(numpy.array([], dtype=numpy.float32)),
                ),
            ],
            None,
        )
    )
    ## tuple → empty tuple
    result = ak.enforce_type(
        ak.to_layout([([1, 2],)], regulararray=False),
        ak.types.from_datashape("()", highlevel=False),
    )
    assert result.layout.is_equal_to(ak.contents.RecordArray([], None, length=1))


def test_list():
    #
    # List types
    result = ak.enforce_type(
        ak.to_layout([[1, 2, 3]]),
        ak.types.from_datashape("var * int64", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.ListOffsetArray(
            ak.index.Index(numpy.array([0, 3], dtype=numpy.int64)),
            ak.contents.NumpyArray(numpy.array([1, 2, 3], dtype=numpy.int64)),
        ),
    )
    result = ak.enforce_type(
        ak.to_layout([[1, 2, 3]]), ak.types.from_datashape("3 * int64", highlevel=False)
    )
    assert result.layout.is_equal_to(
        ak.contents.RegularArray(
            ak.contents.NumpyArray(numpy.array([1, 2, 3], dtype=numpy.int64)), size=3
        ),
    )
    ## Empty list to regular shape
    result = ak.enforce_type(
        ak.to_layout([[]])[:0], ak.types.from_datashape("3 * int64", highlevel=False)
    )
    assert result.layout.is_equal_to(
        ak.contents.RegularArray(
            ak.contents.NumpyArray(numpy.array([], dtype=numpy.int64)), size=3
        ),
    )
    with pytest.raises(ValueError, match=r"converted .* different size"):
        ak.enforce_type(
            ak.to_layout([[1, 2, 3]]),
            ak.types.from_datashape("4 * int64", highlevel=False),
        )

    # Regular types
    result = ak.enforce_type(
        ak.to_regular([[1, 2, 3]], axis=-1, highlevel=False),
        ak.types.from_datashape("var * int64", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.ListOffsetArray(
            ak.index.Index(numpy.array([0, 3], dtype=numpy.int64)),
            ak.contents.NumpyArray(numpy.array([1, 2, 3], dtype=numpy.int64)),
        ),
    )
    result = ak.enforce_type(
        ak.to_regular([[1, 2, 3]], axis=-1, highlevel=False),
        ak.types.from_datashape("3 * int64", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.RegularArray(
            ak.contents.NumpyArray(numpy.array([1, 2, 3], dtype=numpy.int64)), size=3
        ),
    )
    with pytest.raises(ValueError, match=r"different size"):
        ak.enforce_type(
            ak.to_regular([[1, 2, 3]], axis=-1, highlevel=False),
            ak.types.from_datashape("4 * int64", highlevel=False),
        )


def test_option():
    # Options
    ## option → option
    result = ak.enforce_type(
        ak.to_layout([1, None]),
        ak.types.from_datashape("?int64", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.IndexedOptionArray(
            ak.index.Index(numpy.array([0, -1], dtype=numpy.int64)),
            ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
        ),
    )
    ## option → option (packing)
    result = ak.enforce_type(
        ak.to_layout([1, None, 2, 3]),
        ak.types.from_datashape("?float64", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.IndexedOptionArray(
            ak.index.Index(numpy.array([0, -1, 1, 2], dtype=numpy.int64)),
            ak.contents.NumpyArray(numpy.array([1, 2, 3], dtype=numpy.float64)),
        ),
    )

    ## option → no option
    result = ak.enforce_type(
        ak.to_layout([1, None])[:1],
        ak.types.from_datashape("int64", highlevel=False),
    )
    assert ak.almost_equal(
        result, ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64))
    )

    with pytest.raises(ValueError, match=r"if there are no missing values"):
        ak.enforce_type(
            ak.to_layout([1, None]),
            ak.types.from_datashape("int64", highlevel=False),
        )

    ## Add option
    result = ak.enforce_type(
        ak.to_layout([1, 2]),
        ak.types.from_datashape("?int64", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.UnmaskedArray(
            ak.contents.NumpyArray(numpy.array([1, 2], dtype=numpy.int64))
        ),
    )

    ## option[X] → option[unknown]
    layout = ak.to_layout([None, 1, 2, 3])
    result = ak.enforce_type(layout, "?unknown")
    assert result.layout.is_equal_to(
        ak.contents.IndexedOptionArray(
            ak.index.Index64([-1, -1, -1, -1]), ak.contents.EmptyArray()
        )
    )


def test_numpy():
    ## NumPy
    ## 1D → 1D
    result = ak.enforce_type(
        ak.to_layout([1, 2]),
        ak.types.from_datashape("int64", highlevel=False),
    )
    assert ak.almost_equal(result, numpy.array([1, 2], dtype=numpy.int64))

    result = ak.enforce_type(
        ak.to_layout([1, 2]),
        ak.types.from_datashape("float32", highlevel=False),
    )
    assert ak.almost_equal(result, numpy.array([1.0, 2.0], dtype=numpy.float32))

    with pytest.raises(TypeError):
        ak.enforce_type(
            ak.to_layout([1, 2]),
            ak.types.from_datashape("string", highlevel=False),
        )

    ## 1D → 2D
    with pytest.raises(TypeError):
        ak.enforce_type(
            ak.to_layout([1, 2]),
            ak.types.from_datashape("var * int64", highlevel=False),
        )
    with pytest.raises(TypeError):
        ak.enforce_type(
            ak.to_layout([1, 2]),
            ak.types.from_datashape("2 * float32", highlevel=False),
        )
    ## 2D → 1D
    with pytest.raises(TypeError):
        ak.enforce_type(
            ak.to_layout(numpy.zeros((2, 3)), regulararray=False),
            ak.types.from_datashape("int64", highlevel=False),
        )
    with pytest.raises(TypeError):
        ak.enforce_type(
            ak.to_layout(numpy.zeros((2, 3)), regulararray=False),
            ak.types.from_datashape("float32", highlevel=False),
        )

    ## 2D → 2D
    result = ak.enforce_type(
        ak.to_layout(numpy.zeros((2, 3)), regulararray=False),
        ak.types.from_datashape("var * int64", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.ListOffsetArray(
            ak.index.Index(numpy.array([0, 3, 6], dtype=numpy.int64)),
            ak.contents.NumpyArray(numpy.array([0, 0, 0, 0, 0, 0], dtype=numpy.int64)),
        ),
    )

    result = ak.enforce_type(
        ak.to_layout(numpy.zeros((2, 3)), regulararray=False),
        ak.types.from_datashape("3 * float32", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.RegularArray(
            ak.contents.NumpyArray(
                numpy.array([0, 0, 0, 0, 0, 0], dtype=numpy.float32)
            ),
            size=3,
        ),
    )


def test_union():
    # Unions

    ## non union → union
    result = ak.enforce_type(
        ak.to_layout([1, 2]),
        ak.types.from_datashape("union[int64, string]", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.UnionArray(
            tags=ak.index.Index8([0, 0]),
            index=ak.index.Index64([0, 1]),
            contents=[
                ak.contents.NumpyArray(numpy.array([1, 2], dtype=numpy.int64)),
                ak.contents.ListOffsetArray(
                    offsets=ak.index.Index64([0]),
                    content=ak.contents.NumpyArray(
                        numpy.array([], dtype=numpy.uint8),
                        parameters={"__array__": "char"},
                    ),
                    parameters={"__array__": "string"},
                ),
            ],
        )
    )

    with pytest.raises(TypeError):
        ak.enforce_type(
            ak.to_layout([1, 2]),
            ak.types.from_datashape("union[var * int64, string]", highlevel=False),
        )

    ## union → no union (project)
    result = ak.enforce_type(
        # Build union layout, slice to test projection
        ak.to_layout([1, "hi", "bye"])[1:2],
        ak.types.from_datashape("string", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.IndexedArray(
            ak.index.Index64([0]),
            ak.contents.ListOffsetArray(
                offsets=ak.index.Index64([0, 2, 5]),
                content=ak.contents.NumpyArray(
                    numpy.array([104, 105, 98, 121, 101], dtype=numpy.uint8),
                    parameters={"__array__": "char"},
                ),
                parameters={"__array__": "string"},
            ),
        )
    )
    result = ak.enforce_type(
        # Build union layout, slice to test projection
        ak.to_layout([1, "hi", "bye"])[:1],
        ak.types.from_datashape("int64", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.IndexedArray(
            ak.index.Index64([0]),
            ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
        )
    )

    with pytest.raises(TypeError):
        ak.enforce_type(
            ak.to_layout([1, "hi"]),
            ak.types.from_datashape("var * int64", highlevel=False),
        )

    ## union → no union (convert)
    array = ak.concatenate(
        [
            # {x: int64, y: float64}
            [{"x": 1, "y": 2.0}],
            # {y: int64}
            [{"y": 3}],
            # {x: int64, y: float64}
            [{"x": 4, "y": 5.0}],
            # {y: int64}
            [{"y": 6}],
        ],
    )
    assert array.type == ak.types.ArrayType(
        ak.types.UnionType(
            [
                ak.types.RecordType(
                    [ak.types.NumpyType("int64"), ak.types.NumpyType("float64")],
                    ["x", "y"],
                ),
                ak.types.RecordType([ak.types.NumpyType("int64")], ["y"]),
            ]
        ),
        4,
    )
    result = ak.enforce_type(
        array, ak.types.from_datashape("{y: int64}", highlevel=False)
    )
    assert result.layout.is_equal_to(
        ak.contents.IndexedArray(
            index=ak.index.Index64([0, 2, 1, 3]),
            content=ak.contents.RecordArray(
                contents=[
                    ak.contents.NumpyArray(
                        numpy.array([2, 5, 3, 6], dtype=numpy.int64),
                    )
                ],
                fields=["y"],
            ),
        )
    )
    with pytest.raises(ValueError):
        ak.enforce_type(
            ak.to_layout([1, "hi"]), ak.types.from_datashape("string", highlevel=False)
        )

    ## union → same union
    result = ak.enforce_type(
        ak.to_layout([1, "hi"]),
        ak.types.UnionType(
            [
                ak.types.NumpyType("int64"),
                ak.types.ListType(
                    ak.types.NumpyType(
                        "uint8", parameters={"__array__": "char"}, typestr="char"
                    ),
                    parameters={"__array__": "string", "foo": "bar"},
                    typestr="string",
                ),
            ]
        ),
    )
    assert result.layout.is_equal_to(
        ak.contents.UnionArray(
            tags=ak.index.Index8([0, 1]),
            index=ak.index.Index64([0, 0]),
            contents=[
                ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
                ak.contents.ListOffsetArray(
                    offsets=ak.index.Index64([0, 2]),
                    content=ak.contents.NumpyArray(
                        numpy.array([104, 105], dtype=numpy.uint8),
                        parameters={"__array__": "char"},
                    ),
                    parameters={"__array__": "string", "foo": "bar"},
                ),
            ],
        )
    )

    ## union → bigger union
    result = ak.enforce_type(
        ak.to_layout([1, "hi"]),
        ak.types.from_datashape("union[int64, string, datetime64]", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.UnionArray(
            tags=ak.index.Index8([0, 1]),
            index=ak.index.Index64([0, 0]),
            contents=[
                ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
                ak.contents.ListOffsetArray(
                    offsets=ak.index.Index64([0, 2]),
                    content=ak.contents.NumpyArray(
                        numpy.array([104, 105], dtype=numpy.uint8),
                        parameters={"__array__": "char"},
                    ),
                    parameters={"__array__": "string"},
                ),
                ak.contents.NumpyArray(numpy.array([], dtype=numpy.datetime64)),
            ],
        )
    )

    ## union → different union (same N)
    result = ak.enforce_type(
        ak.to_layout([1, "hi"]),
        ak.types.UnionType(
            [
                ak.types.NumpyType("float32"),
                ak.types.ListType(
                    ak.types.NumpyType(
                        "uint8", parameters={"__array__": "char"}, typestr="char"
                    ),
                    parameters={"__array__": "string", "foo": "bar"},
                    typestr="string",
                ),
            ]
        ),
    )
    assert result.layout.is_equal_to(
        ak.contents.UnionArray(
            tags=ak.index.Index8([0, 1]),
            index=ak.index.Index64([0, 0]),
            contents=[
                ak.contents.NumpyArray(numpy.array([1], dtype=numpy.float32)),
                ak.contents.ListOffsetArray(
                    offsets=ak.index.Index64([0, 2]),
                    content=ak.contents.NumpyArray(
                        numpy.array([104, 105], dtype=numpy.uint8),
                        parameters={"__array__": "char"},
                    ),
                    parameters={"__array__": "string", "foo": "bar"},
                ),
            ],
        )
    )

    ## union → different union (smaller N)
    result = ak.enforce_type(
        ak.to_layout([1, "hi", [1j, 2j]])[:2],
        "union[int64, string]",
    )
    assert result.layout.is_equal_to(
        ak.contents.UnionArray(
            tags=ak.index.Index8([0, 1]),
            index=ak.index.Index64([0, 0]),
            contents=[
                ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
                ak.contents.ListOffsetArray(
                    offsets=ak.index.Index64([0, 2]),
                    content=ak.contents.NumpyArray(
                        numpy.array([104, 105], dtype=numpy.uint8),
                        parameters={"__array__": "char"},
                    ),
                    parameters={"__array__": "string"},
                ),
            ],
        )
    )

    ## union → incompatible different union (same N)
    with pytest.raises(TypeError):
        ak.enforce_type(
            ak.to_layout([1, "hi"]),
            ak.types.from_datashape("union[int64, bool]", highlevel=False),
        )

    ## union → different union (same N, more than one change)
    with pytest.raises(TypeError):
        ak.enforce_type(
            ak.to_layout([1, "hi", False]),
            ak.types.from_datashape(
                "union[datetime64, string, float32]", highlevel=False
            ),
        )

    ## union of union → union of extended union
    layout = ak.contents.UnionArray(
        ak.index.Index8([0, 1]),
        ak.index.Index64([0, 0]),
        [
            ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64([0, 2]),
                ak.contents.UnionArray(
                    ak.index.Index8([0, 1]),
                    ak.index.Index64([0, 0]),
                    [
                        ak.contents.NumpyArray(numpy.array([2], dtype=numpy.int64)),
                        ak.contents.ListOffsetArray(
                            ak.index.Index64([0, 2]),
                            ak.contents.RecordArray(
                                [
                                    ak.contents.IndexedOptionArray(
                                        ak.index.Index64([0, -1]),
                                        ak.contents.NumpyArray(
                                            numpy.array([1], dtype=numpy.int64)
                                        ),
                                    ),
                                    ak.contents.IndexedOptionArray(
                                        ak.index.Index64([-1, 0]),
                                        ak.contents.NumpyArray(
                                            numpy.array([2], dtype=numpy.int64)
                                        ),
                                    ),
                                ],
                                ["x", "y"],
                            ),
                        ),
                    ],
                ),
            ),
        ],
    )
    result = ak.enforce_type(
        layout,
        """
    union[
        int64,
        var * union[
            int64,
            var * {
                x: ?int64,
                y: ?int64,
                z: ?string
            }
        ]
    ]
                             """,
    )
    assert result.layout.is_equal_to(
        ak.contents.UnionArray(
            ak.index.Index8([0, 1]),
            ak.index.Index64([0, 0]),
            [
                ak.contents.NumpyArray(numpy.array([1], dtype=numpy.int64)),
                ak.contents.ListOffsetArray(
                    ak.index.Index64([0, 2]),
                    ak.contents.UnionArray(
                        ak.index.Index8([0, 1]),
                        ak.index.Index64([0, 0]),
                        [
                            ak.contents.NumpyArray(numpy.array([2], dtype=numpy.int64)),
                            ak.contents.ListOffsetArray(
                                ak.index.Index64([0, 2]),
                                ak.contents.RecordArray(
                                    [
                                        ak.contents.IndexedOptionArray(
                                            ak.index.Index64([0, -1]),
                                            ak.contents.NumpyArray(
                                                numpy.array([1], dtype=numpy.int64)
                                            ),
                                        ),
                                        ak.contents.IndexedOptionArray(
                                            ak.index.Index64([-1, 0]),
                                            ak.contents.NumpyArray(
                                                numpy.array([2], dtype=numpy.int64)
                                            ),
                                        ),
                                        ak.contents.IndexedOptionArray(
                                            ak.index.Index64([-1, -1]),
                                            ak.contents.ListOffsetArray(
                                                ak.index.Index64([0]),
                                                ak.contents.NumpyArray(
                                                    numpy.empty(0, dtype=numpy.uint8),
                                                    parameters={"__array__": "char"},
                                                ),
                                                parameters={"__array__": "string"},
                                            ),
                                        ),
                                    ],
                                    ["x", "y", "z"],
                                ),
                            ),
                        ],
                    ),
                ),
            ],
        )
    )


def test_string():
    ## string -> bytestring
    result = ak.enforce_type(
        ak.to_layout(["hello world"]),
        ak.types.from_datashape("bytes", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.ListOffsetArray(
            offsets=ak.index.Index64([0, 11]),
            content=ak.contents.NumpyArray(
                numpy.array(
                    [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100],
                    dtype=numpy.uint8,
                ),
                parameters={"__array__": "byte"},
            ),
            parameters={"__array__": "bytestring"},
        ),
    )

    ## bytestring -> string
    result = ak.enforce_type(
        ak.to_layout([b"hello world"]),
        ak.types.from_datashape("string", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.ListOffsetArray(
            offsets=ak.index.Index64([0, 11]),
            content=ak.contents.NumpyArray(
                numpy.array(
                    [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100],
                    dtype=numpy.uint8,
                ),
                parameters={"__array__": "char"},
            ),
            parameters={"__array__": "string"},
        ),
    )

    ## string -> string
    result = ak.enforce_type(
        ak.to_layout(["hello world"]),
        ak.types.from_datashape("string", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.ListOffsetArray(
            offsets=ak.index.Index64([0, 11]),
            content=ak.contents.NumpyArray(
                numpy.array(
                    [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100],
                    dtype=numpy.uint8,
                ),
                parameters={"__array__": "char"},
            ),
            parameters={"__array__": "string"},
        ),
    )

    ## bytestring -> bytestring
    result = ak.enforce_type(
        ak.to_layout([b"hello world"]),
        ak.types.from_datashape("bytes", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.ListOffsetArray(
            offsets=ak.index.Index64([0, 11]),
            content=ak.contents.NumpyArray(
                numpy.array(
                    [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100],
                    dtype=numpy.uint8,
                ),
                parameters={"__array__": "byte"},
            ),
            parameters={"__array__": "bytestring"},
        ),
    )

    ## bytestring -> list of byte
    result = ak.enforce_type(
        ak.to_layout([b"hello world"]),
        ak.types.from_datashape("var * byte", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.ListOffsetArray(
            offsets=ak.index.Index64([0, 11]),
            content=ak.contents.NumpyArray(
                numpy.array(
                    [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100],
                    dtype=numpy.uint8,
                ),
                parameters={"__array__": "byte"},
            ),
        ),
    )

    ## bytestring -> list of int64
    result = ak.enforce_type(
        ak.to_layout([b"hello world"]),
        ak.types.from_datashape("var * int64", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.ListOffsetArray(
            offsets=ak.index.Index64([0, 11]),
            content=ak.contents.NumpyArray(
                numpy.array(
                    [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100],
                    dtype=numpy.int64,
                )
            ),
        ),
    )

    ## list of int64 -> string
    result = ak.enforce_type(
        ak.without_parameters([b"hello world"]),
        ak.types.from_datashape("string", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.ListOffsetArray(
            offsets=ak.index.Index64([0, 11]),
            content=ak.contents.NumpyArray(
                numpy.array(
                    [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100],
                    dtype=numpy.uint8,
                ),
                parameters={"__array__": "char"},
            ),
            parameters={"__array__": "string"},
        ),
    )


def test_highlevel():
    with pytest.raises(TypeError, match=r"High-level type objects are not supported"):
        ak.enforce_type(
            ak.to_layout(["hello world"]),
            ak.types.from_datashape("1 * bytes"),
        )
    with pytest.raises(TypeError, match=r"High-level type objects are not supported"):
        ak.enforce_type(
            ak.to_layout([{"msg": "hello world"}]),
            ak.types.ScalarType(
                ak.types.RecordType(
                    [
                        ak.types.ListType(
                            ak.types.NumpyType(
                                "uint8", parameters={"__array__": "char"}
                            ),
                            parameters={"__array__": "string"},
                        )
                    ],
                    ["msg"],
                )
            ),
        )


def test_single_record():
    result = ak.enforce_type(
        ak.Record({"x": [1, 2]}),
        ak.types.from_datashape("{x: var * float64}", highlevel=False),
        highlevel=False,
    )
    assert isinstance(result, ak.record.Record)
    assert ak.almost_equal(
        result.array,
        ak.contents.RecordArray(
            [
                ak.contents.ListOffsetArray(
                    ak.index.Index(numpy.array([0, 2], dtype=numpy.int64)),
                    ak.contents.NumpyArray(numpy.array([1, 2], dtype=numpy.float64)),
                )
            ],
            ["x"],
        ),
    )


def test_indexed():
    # Non-packing (because dtype hasn't changed)
    result = ak.enforce_type(
        ak.contents.IndexedArray(
            ak.index.Index64([0, 2]),
            ak.contents.ListOffsetArray(
                ak.index.Index64([0, 3, 6, 9]),
                ak.contents.NumpyArray(numpy.arange(9, dtype=numpy.int64)),
            ),
        ),
        ak.types.from_datashape(
            'var * int64[parameters={"key": "value"}]', highlevel=False
        ),
    )
    assert result.layout.is_equal_to(
        ak.contents.IndexedArray(
            ak.index.Index64([0, 2]),
            ak.contents.ListOffsetArray(
                ak.index.Index(numpy.array([0, 3, 6, 9], dtype=numpy.int64)),
                ak.contents.NumpyArray(
                    numpy.array([0, 1, 2, 3, 4, 5, 6, 7, 8], dtype=numpy.int64),
                    parameters={"key": "value"},
                ),
            ),
        )
    )
    # Packing
    result = ak.enforce_type(
        ak.contents.IndexedArray(
            ak.index.Index64([0, 2]),
            ak.contents.ListOffsetArray(
                ak.index.Index64([0, 3, 6, 9]),
                ak.contents.NumpyArray(numpy.arange(9, dtype=numpy.int64)),
            ),
        ),
        ak.types.from_datashape("var * float32", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.ListOffsetArray(
            ak.index.Index64([0, 3, 6]),
            ak.contents.NumpyArray(
                numpy.array([0, 1, 2, 6, 7, 8], dtype=numpy.float32)
            ),
        )
    )


def test_unknown():
    # unknown → unknown
    layout = ak.contents.EmptyArray()
    assert ak.enforce_type(layout, "unknown").layout.is_equal_to(layout)

    # unknown → other
    layout = ak.contents.EmptyArray()
    assert ak.enforce_type(layout, "int64").layout.is_equal_to(
        ak.contents.NumpyArray(numpy.empty(0, numpy.int64))
    )

    # other → unknown
    with pytest.raises(
        TypeError, match=r"cannot convert non-EmptyArray layouts to a bare UnknownType"
    ):
        layout = ak.contents.NumpyArray(numpy.empty(0, numpy.int64))
        ak.enforce_type(layout, "unknown")

    # unknown → other
    layout = ak.contents.NumpyArray(numpy.empty(0, numpy.int64))
    assert ak.enforce_type(layout, "?unknown").layout.is_equal_to(
        ak.contents.IndexedOptionArray(ak.index.Index64([]), ak.contents.EmptyArray())
    )


def test_misc():
    # These tests ensure good coverage over our helper functions
    ## option → option (inside indexed)
    layout = ak.to_layout([{"x": [1, 2, None]}, None, {"x": [3, 4, None]}])[[0, 2], :2]
    result = ak.enforce_type(
        layout,
        """
    ?{
        x: var * ?float32
    }
        """,
    )
    assert result.layout.is_equal_to(
        ak.contents.IndexedOptionArray(
            ak.index.Index64([0, 1]),
            ak.contents.RecordArray(
                [
                    ak.contents.ListOffsetArray(
                        ak.index.Index64([0, 2, 4]),
                        ak.contents.IndexedOptionArray(
                            ak.index.Index64([0, 1, 2, 3]),
                            ak.contents.NumpyArray(
                                numpy.array([1, 2, 3, 4], dtype=numpy.float32)
                            ),
                        ),
                    )
                ],
                ["x"],
            ),
        )
    )
    ## no option → option (inside indexed)
    layout = ak.to_layout(
        [
            {"x": [1, 2, None]},
            {"x": [9, 9, None]},
            {"x": [3, 4, None]},
            {"x": [8, 8, None]},
        ]
    )[[0, 2], :2]
    result = ak.enforce_type(
        layout,
        """
    ?{
        x: var * ?float32
    }
        """,
    )
    assert result.layout.is_equal_to(
        ak.contents.UnmaskedArray(
            ak.contents.RecordArray(
                [
                    ak.contents.ListOffsetArray(
                        ak.index.Index64([0, 2, 4]),
                        ak.contents.IndexedOptionArray(
                            ak.index.Index64([0, 1, 2, 3]),
                            ak.contents.NumpyArray(
                                numpy.array([1, 2, 3, 4], dtype=numpy.float32)
                            ),
                        ),
                    )
                ],
                ["x"],
            ),
        )
    )

    ## option (indexed) list of union → option list of no union (project)
    layout = ak.to_layout([[1, "hi", "bye"], None])[[0, 1], 1:2]
    assert isinstance(layout, ak.contents.IndexedOptionArray)
    result = ak.enforce_type(
        # Build union layout, slice to test projection/no-projection
        # wrap union in outer option-of-list, and index it to produce a IndexedOptionArray (to test for packing)
        layout,
        ak.types.from_datashape("?var * string", highlevel=False),
    )
    assert result.layout.is_equal_to(
        ak.contents.IndexedOptionArray(
            ak.index.Index64([0, -1]),
            ak.contents.ListOffsetArray(
                ak.index.Index64([0, 1]),
                # Indexed type because the string dtype doesn't change, so we don't need to back below this point
                ak.contents.IndexedArray(
                    ak.index.Index64([0]),
                    ak.contents.ListOffsetArray(
                        offsets=ak.index.Index64([0, 2, 5]),
                        content=ak.contents.NumpyArray(
                            numpy.array([104, 105, 98, 121, 101], dtype=numpy.uint8),
                            parameters={"__array__": "char"},
                        ),
                        parameters={"__array__": "string"},
                    ),
                ),
            ),
        ),
    )
